'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 
import numpy as np
import scipy.sparse.linalg as spl


class Laplacian1D():
    '''
    The Laplace operator.
    '''
    def __init__(self,Mesh,mu,assemble='no'):
        self.mu   = mu
        self.Mesh = Mesh
        self.g11 = self.geometric()
        if assemble == 'no':
            self.matrix=spl.LinearOperator( (self.Mesh.dof,self.Mesh.dof),\
                                            matvec=self.action_python,dtype='float')    
        #if assemble == 'yes':
    
    def geometric(self):
        w = np.zeros((self.Mesh.noe,self.Mesh.n),float)
        g11=1.0/self.Mesh.jac
        return g11
        
    def action_python(self,v_in):
        w = np.zeros((self.Mesh.noe,self.Mesh.n),float)
        u = self.Mesh.mapping_q(v_in)
        for k in range(self.Mesh.noe):
            for m in range(self.Mesh.n):
                for n in range(self.Mesh.n):
                    for alpha in range(self.Mesh.n):
                        w[k,m]=w[k,m]+self.Mesh.weights[alpha]*self.g11[k,alpha]\
                                     *self.mu[k,alpha]\
                                     *self.Mesh.D[alpha,n]*self.Mesh.D[alpha,m]*u[k,n]
        # Mapping
        v_out=self.Mesh.mapping_qt(w)
        # Mask
        v_out = self.Mesh.mask(v_out)
        return v_out
    
#    def action(self,v):
#        # Mapping from global to local
#        u = self.Mesh.mapping_Q(v)
#        # Matrix-vector product
#        w = fortrancode.action_of_laplacian(u,self.mu,self.Mesh.weights,self.Mesh.jac,self.Mesh.D)
#        # Mapping from local to global
#        w_out = self.Mesh.mapping_QT(w)
#        # Mask
#        w_out = self.Mesh.mask(w_out)
#        return w_out
        
#    def action_assemble(self):
#        print 'assemble sem...'
#        nz = self.Mesh.noe * self.Mesh.n * self.Mesh.n
#
#        weights = self.Mesh.weights
#        D   = self.Mesh.D
#        mu  = self.mu
#        g11 = self.g11
#        
#        nz_b=self.Mesh.noe * self.Mesh.n * self.Mesh.n
#        irow,jcolumn,value = fortrancode.assemble_full(weights,mu,g11,D,self.Mesh.theta,nz_b)
#        A = pysparse.spmatrix.ll_mat(self.Mesh.dof, self.Mesh.dof, nz)
#        for k in range(len(value)):
#            i=int(irow[k])
#            j=int(jcolumn[k])
#            A[i,j]=A[i,j]+value[k]
#        # Mask
#        for k in range(self.Mesh.noe_bc):
#            [i, phys, bc_type, element_edge]=self.Mesh.theta_phys_bc(k)
#            if bc_type == "Dir":
#                A[int(i),:] = 0.0
#        t2 = time.clock()
#        print '...assemble time=',t2-t1
#        print 'A.nz=',A.nnz
#        print '...assemble sem'
#        return A 

class Mass1D():
    '''
    The mass matrix.
    
    :param Mesh: An object of one of the mesh classes
    
    :attributes: *matrix* - Returns the mass matrix :math:`M\in \\mathbb R^{dof\\times dof}`. Since 
                              this matrix is diagonal it is always assembled. 
    
    Usage::
    
      import sempy
      
      B=Mass(mesh).matrix
      
    :todo: * inverse mass matrix.
      
    '''
    def __init__(self,Mesh):
        self.Mesh = Mesh
        self.matrix = spl.LinearOperator( (self.Mesh.dof,self.Mesh.dof),\
                                           matvec=self.action,dtype='float') 
        
    def action(self,u):
        w=np.zeros((self.Mesh.noe,self.Mesh.n),float)
        u_local = self.Mesh.mapping_Q(u)
        for k in range(self.Mesh.noe):
            for m in range(self.Mesh.n):
                w[k,m] = self.Mesh.weights[m] * self.Mesh.jac[k,m] * u_local[k,m]
        # Gather
        w_out = self.Mesh.mapping_qt(w)
        # Mask
        w_out = self.Mesh.mask(w_out)
        return w_out
        
    def action_of_matrix(self,f):
        '''
        Action of the mass matrix on a set of basis coefficients given in local data
        representation. 
        
        :param f: set of basis coefficients given in local data representation. 
        
        This is useful for functions in :math:`L^2(\\Omega)`
        
        Usage::
        
          import sempy
          
          f = np.ones((mesh.noe,mesh.n,mesh.n),float)
          F = Mass(mesh).action_of_matrix(f)
          
        '''
        w=np.zeros((self.Mesh.noe,self.Mesh.n),float)
        for k in range(self.Mesh.noe):
            for m in range(self.Mesh.n):
                w[k,m] = self.Mesh.weights[m] * self.Mesh.jac[k,m] * f[k,m]
        # Gather
        w_out = self.Mesh.mapping_qt(w)
        # Mask
        w_out = self.Mesh.mask(w_out)
        return w_out



'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 

import numpy as np
import scipy as sp
import scipy.sparse.linalg as spl
import pickle as pickle
import scipy.io as sio
import sempy
import ode_f90
# Compatibility with older versions
import pysparse
if int(pysparse.__version__[2]) >= 2:
    from pysparse.sparse.pysparseMatrix import PysparseMatrix #as pysparseMatrix
    from pysparse import direct
    
if int(pysparse.__version__[2]) <=1:
    from pysparse.pysparseMatrix import PysparseMatrix

class Time:
    '''
    Sets up the temporal domain. 
       
    >>> import sempy
    >>> Y = sempy.ode.Time(start_time=0,end_time=10,time_steps=10)


    :param float start_time: Start time 
    :param float end_time: End time 
    :param int time_steps: Number of timesteps

    :attributes: * **h** (float)- Time step size
                 * **time** (float)- Time
    '''
    def __init__( self, Space, start_time=0, end_time=1.0, 
                  time_steps = 2):
        self.Space = Space
        self.start_time = float(start_time)
        self.end_time   = float(end_time)
        self.time_steps = time_steps
        self.h          = (self.end_time-self.start_time)/float(self.time_steps)
        self.time       = float(start_time)
        #solver = sempy.linsolvers.Krylov()
   
    def increment(self):
        """
        Increments the :literal:`time` attribute with the time step size 
        :literal:`h`. 

        Usage::
        
           import sempy
        
           Y = sempy.ode.Time( start_time=0,end_time=10,time_steps=10)
           Y.increment()

        """
        self.time += self.h
        
       
def dirk111():
    '''
    Euler method. Forward/backward.
    '''
    a,b,c    =np.zeros((2,2)),np.zeros(2),np.zeros(2)#float
    ah,bh,ch =np.zeros((2,2)),np.zeros(2),np.zeros(2)#float
    a[0,0],  a[0,1]=0.0, 0.0
    a[1,0],  a[1,1]=0.0, 1.0
    ah[0,0],ah[0,1]=0.0, 0.0
    ah[1,0],ah[1,1]=1.0, 0.0
    b[0],b[1]  =0.0,1.0 
    bh[0],bh[1]=1.0,0.0 
    c[0],c[1]  =0.0,1.0 
    ch[0],ch[1]=0.0,1.0 
    stages=2
    return a,b,c,ah,bh,ch,stages 

def dirk122():
    '''
    PRK method.
    '''
    a,b,c    =np.zeros((2,2),float),np.zeros((2),float),np.zeros((2),float)
    ah,bh,ch =np.zeros((2,2),float),np.zeros((2),float),np.zeros((2),float)
    a[0,0],  a[0,1]=0.0, 0.0
    a[1,0],  a[1,1]=0.0, 0.5
    ah[0,0],ah[0,1]=0.0, 0.0
    ah[1,0],ah[1,1]=0.5, 0.0
    b[0],b[1]  =0.0,1.0 
    bh[0],bh[1]=0.0,1.0 
    c[0],c[1]  =0.0,0.5 
    ch[0],ch[1]=0.0,0.5 
    stages=2
    return a,b,c,ah,bh,ch,stages 
    
def dirk222():
    '''
    PRK.
    '''
    a,b,c    =np.zeros((3,3),float),np.zeros((3),float),np.zeros((3),float)
    ah,bh,ch =np.zeros((3,3),float),np.zeros((3),float),np.zeros((3),float)
    gamma = (2.0 - sp.sqrt(2.0))/2.0
    delta = 1.0 - 1.0/(2.0*gamma)
    a[0,0], a[0,1], a[0,2] = 0.0, 0.0      ,0.0
    a[1,0], a[1,1], a[1,2] = 0.0, gamma    ,0.0
    a[2,0], a[2,1], a[2,2] = 0.0, 1.0-gamma,gamma

    ah[0,0], ah[0,1], ah[0,2] = 0.0,   0.0      ,0.0
    ah[1,0], ah[1,1], ah[1,2] = gamma, 0.0      ,0.0
    ah[2,0], ah[2,1], ah[2,2] = delta, 1.0-delta,0.0

    b[0], b[1], b[2]    = 0.0,1.0-gamma,gamma
    bh[0], bh[1], bh[2] = delta, 1.0-delta,0.0
    c[0], c[1], c[2]    = 0.0,gamma,1.0
    ch[0], ch[1], ch[2] = 0.0,gamma,1.0
    stages=3
    return a,b,c,ah,bh,ch,stages 

def irk2():
    a,b,c    = np.zeros((2,2),float),np.zeros((2),float),np.zeros((2),float)
    ah,bh,ch = np.zeros((2,2),float),np.zeros((2),float),np.zeros((2),float)
    gamma    = (2.0 - np.sqrt(2.0))/2.0
    #delta=1.0-1.0/(2.0*gamma)
    a[0,0],  a[0,1]= gamma    , 0.0
    a[1,0],  a[1,1]= 1.0-gamma, gamma
    b[0],b[1]      = 1.0-gamma, gamma
    c[0],c[1]      = gamma, 1.0
    stages=2
    return a,b,c,ah,bh,ch,stages 

def dirk443():
    '''
    PRK method of second order.
    '''
    a,b,c    =np.zeros((5,5)),np.zeros(5),np.zeros(5)#float
    ah,bh,ch =np.zeros((5,5)),np.zeros(5),np.zeros(5)#float
    gamma=(2.0-sp.sqrt(2.0))/2.0
    delta=1.0-1.0/(2.0*gamma)
     
    a[1,1]=1.0/2.0
    a[2,1],a[2,2]=1.0/6.0,1.0/2.0
    a[3,1],a[3,2],a[3,3]=-1.0/2.0,1.0/2.0,1.0/2.0
    a[4,1],a[4,2],a[4,3],a[4,4]=3.0/2.0,-3.0/2.0,1.0/2.0,1.0/2.0
    b[1],b[2],b[3],b[4]=3.0/2.0,-3.0/2.0,1.0/2.0,1.0/2.0
    c[1],c[2],c[3],c[4]=1.0/2.0,2.0/3.0,1.0/2.0,1.0

    ah[1,0] =1.0/2.0
    ah[2,0],ah[2,1]=11.0/18.0,1.0/18.0  
    ah[3,0],ah[3,1],ah[3,2]=5.0/6.0,-5.0/6.0,1.0/2.0
    ah[4,0],ah[4,1],ah[4,2],ah[4,3]=1.0/4.0,7.0/4.0,3.0/4.0,-7.0/4.0
    ch[1],ch[2],ch[3],ch[4]=1.0/2.0,2.0/3.0,1.0/2.0,1.0
    bh[0],bh[1],bh[2],bh[3]=ah[4,0],ah[4,1],ah[4,2],ah[4,3]

    stages=5
    return a,b,c,ah,bh,ch,stages 

def dirk252():
    '''
    PRK method of second order.
    '''
    a,b,c    =np.zeros((5,5)),np.zeros(5),np.zeros(5)#float
    ah,bh,ch =np.zeros((5,5)),np.zeros(5),np.zeros(5)#float
    gamma=(2.0-sp.sqrt(2.0))/2.0
    delta=1.0-1.0/(2.0*gamma)
     
    a[3,3]=gamma
    a[4,3]=1.0-gamma
    a[4,4]=gamma
    b[3]=1.0-gamma
    b[4]=gamma
    c[3]=gamma
    c[4]=1.0

    ah[1,0]=1.0/4.0
    ah[2,1]=1.0/3.0
    ah[3,2]=gamma
    ah[4,3]=1.0/(2.0*gamma)
    ah[4,0]=-np.sqrt(1.0/2.0)
    ch[1]=ah[1,0]
    ch[2]=ah[2,1]
    ch[3]=ah[3,2]
    ch[4]=1.0
    bh[0]=ah[4,0]
    bh[3]=ah[4,3]

    stages=5
    return a,b,c,ah,bh,ch,stages 

def erk4():
    '''
    Explicit RK method of order 4.
    '''
    a,b,c    =np.zeros((4,4),float),np.zeros((4),float),np.zeros((4),float)
    ah,bh,ch =np.zeros((4,4),float),np.zeros((4),float),np.zeros((4),float)

    ah[1,0]=1.0/2.0
    ah[2,1]=1.0/2.0
    ah[3,2]=1.0

    ch[1],ch[2],ch[3]=ah[1,0],ah[2,1],ah[3,2]
    bh[0],bh[1],bh[2],bh[3]=1.0/6.0 ,1.0/3.0, 1.0/3.0, 1.0/6.0
    stages=4
    return a,b,c,ah,bh,ch,stages 

def __emptyfunc__(y,t):
    v=np.zeros((y.size))
    return v

def __pre__(y):
    return y

def __bc__(y,t):
    return y

class PRK():
    '''
    Implementation of partitioned Runge-Kutta (PRK) method
    for the solution of ODE systems on the form:  
    
    .. math::
    
       M\\frac{dy}{dt}=Ly+N(y,t)
    
    :param y: Initial condition
    :type y: A collection of :class:`sempy.Function`
    :param Time: An instance of the time class.
    :type Time: :class:`sempy.ode.Time`
    :kwargs: * **L** - Linear operator
             * **L_pre** - The FEM equivalent of L. Used to create 
               preconditioner. 
             * **force_function** - A function, possibly non-linear. 
               This is the function :math:`N(y,t)` in the equation above. 
               Called as:: 
               
                 F=force_function(y,t)
                 
             * **bound_cond** - Strong BCs. Called as:: 
               
                 y=bound_cond(y,t)
                 
             * **prk_method** (*string*) - The PRK method to be used. 
               The options are:
                    * :literal:`dirk111`
                    * :literal:`dirk122`
                    * :literal:`dirk222`
                    * :literal:`dirk443`
             * **linear_solver** - The linear solver. The default is 
               CG. If L is not SPD, then one can for example use bicgstab.
             * **iter_tol** (*float*) - Iterative tolerance.
             * **maxiter** (*int*) - Maximum number of iterations.
    '''
    def __init__( self, y, Time, L = None, L_pre= None,
                  force_function = None, bound_cond = __bc__,
                  prk_method = None, file_increment = 1000,
                  hyper_filter = False,call_back=None,
                  linear_solver = None, iter_tol = 1e-10,
                  maxiter = 1000,silent=False):
        # Solution vector
        self.number_of_eq = len(y)
        self.y0 = self.output_vector(y,adopt_filename=True)
        self.y_help = self.output_vector(y)
        self.y_out = self.output_vector(y)
        self.hyper_filter = hyper_filter
        self.call_back = call_back 
        self.silent = silent

        self.file_increment= file_increment
        # Degrees of freedom
        dof = 0
        for i in range(self.number_of_eq):
            dof = dof + y[i].Space.dof 
        self.vector_size = dof
        # Strong BC
        self.bound_cond = bound_cond
        # Temporal domain
        self.Time = Time
        # PRK method
        self.prk_method = prk_method
        [self.a,self.b,self.c,self.ah,self.bh,self.ch,self.stages]=\
               self.runge_kutta_method(self.prk_method)
        # Linear operator
        self.L_in = L
        if self.L_in == None:
            self.L_in = []
            G=[]
            for i in range( self.number_of_eq ):
                G.append(0)
            for i in range( self.number_of_eq ):
                self.L_in.append(G) 
            del G

        self.L = spl.LinearOperator( (self.vector_size,
                                      self.vector_size),
                                      matvec = self.__action_linear_operator,
                                      dtype = 'float')   
        # Mass matrix
        self.M, self.M_inv = self.mass_matrix()
        # Preconditioner
        self.L_pre = L_pre
        if self.L_pre == None:
            self.L_pre_inv = 'none'
        else:
            self.L_pre_inv = self.__preconditioner(self.L_pre)
        # Non-linear function
        self.force_function = force_function
        # Linear solver:    
        self.linear_solver = linear_solver
        self.iter_tol = iter_tol
        self.maxiter=maxiter
        if self.linear_solver == None:
            self.linear_solver = sempy.linsolvers.Krylov(
                                                tol = self.iter_tol,
                                                solver_type = 'cg',
                                                maxiter=self.maxiter,
                                                pre=self.L_pre_inv)
    def to_file(self,y_vec):
        '''
        Print to file.
        '''
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + self.y0[i].Space.dof
            temp = self.y0[i].Space.mapping_q( y_vec[dof_1:dof_2] )
            self.y0[i].basis_coeff = np.copy(temp)
            
        for i in range(self.number_of_eq):
            if not self.y0[i].filename == 'none':
                self.y0[i].to_file()
        
        
    def solve(self):
        '''
        Solves the ODE system. 
        '''
        #-- Solution vector y: y[0]=y^{n+1}, y[1]=y^{n}, y[2]=y^{n-1}, 
        # and so on
        ystart = np.zeros( (self.vector_size), float )
        yend   = np.zeros( (self.vector_size), float )
        ystart[:] = self.component_to_vector( self.y0 )
        # Timesteps to file:
        k_file = range( self.file_increment, 
                        self.Time.time_steps+self.file_increment, 
                        self.file_increment )
        #-To file
        #self.to_file(ystart)
        # -- Callback
        if not self.call_back == None:
            y_out = self.vector_to_component(ystart)  
            self.call_back(y_out ,self.Time.time,0)
        # --Checking for b_i=a_si
        alpha=0.0
        for i in range(self.stages):
            alpha = alpha + np.abs(self.b[i]-self.a[self.stages-1,i]) + \
                            np.abs(self.bh[i]-self.ah[self.stages-1,i])
        #-- Time-stepping
        for i in range(self.Time.time_steps):
            if not self.silent:
                print '--> time step no=',i+1
            yend = self.__timestepping__(ystart,alpha)
            # Hyper filter
            #if self.hyper_filter:
            #    y_out = self.vector_to_component(yend)
            #    for k in range(self.number_of_eq):
            #        y_out[k].hyper_filter()
            #    #y_out = self.bound_cond( y_out,self.Time.time +self.Time.h)
            #    yend = self.component_to_vector(y_out)
                
            self.Time.increment()
            ystart = np.copy(yend)
            # -- Callback
            if not self.call_back == None:
                y_out = self.vector_to_component(yend)  
                self.call_back(y_out ,self.Time.time,i+1)
            # To file
            #if (i+1) in k_file:
            #    self.to_file(ystart)
            if not self.silent:
                print 'time=',self.Time.time
            
        y_out = self.vector_to_component(yend)    
        return y_out
        
    def __timestepping__(self,ystart,alpha):
        '''
        Advancing the solution one time-step.
        '''
        #-- Solution vector yi
        yi = np.zeros((self.stages,self.vector_size),float)
        fi = np.zeros((self.stages,self.vector_size),float)
        F  = np.zeros((self.vector_size),float)
        b  = np.zeros((self.vector_size),float)
        
        #for i in range(self.stages):
        #    yi[i,:] = np.copy(ystart[:])
        yi[0,:] = np.copy(ystart[:])
            
        #-- Computing the y_i's
        for i in range(self.stages):
            F[:] = 0.0
            for j in range(i):
                # Explicit
                if np.abs(self.ah[i,j]) > 1.0e-10:
                    #F[:] = F[:] + self.Time.h * self.ah[i,j] * fi[j][:]
                    F[:] = ode_f90.vec_add(F[:],self.Time.h * self.ah[i,j],
                                           fi[j][:])
                # Implicit
                if np.abs(self.a[i,j]) > 1.0e-10:
                    b[:] = self.L*yi[j][:]
                    #F[:] = F[:] + self.Time.h*self.a[i,j]*b[:]
                    F[:] = ode_f90.vec_add(F[:],self.Time.h*self.a[i,j],b[:])
                    # include weak BCs in this step. 
                    
            b[:] = self.M * ystart[:] + F[:]
            if np.abs(self.a[i,i]) > 0.0:
                # System solve required. Solve the system Hy[i]=b
                local_time  = self.Time.time + self.c[i] * self.Time.h
                yi[i][:] = self.apply_bc(yi[i][:],local_time)
                H = sempy.operators.MultipleOperators([self.M,self.L],\
                       scaling_factor=[1.0, -self.Time.h*self.a[i,i]],\
                       assemble='no').matrix
                [yi[i][:],flag] = self.linear_solver.solve(H,b,yi[i][:])
                if not self.silent:
                    print 'iter=',flag
                # 1) Include optional hyperbolic filter here. 
            else:
                # System solve not required
                time  = self.Time.time + self.ch[i] * self.Time.h
                yi[i][:] = self.M_inv*b[:]
                #yi[i][:] = self.bound_cond(yi[i][:],time)
                yi[i][:] = self.apply_bc(yi[i][:],time)
            time  = self.Time.time + self.ch[i] * self.Time.h
            fi[i][:] = self.force_vec(yi[i][:],time)
        # Computing the final value yend:
        if np.abs(alpha) > 1.0e-10 :
            if not self.silent:
                print 'last step required'
            F[:] = 0.0
            for i in range(self.stages):
                if np.abs(self.bh[i]) > 0.0:
                    #F[:] = F[:] + self.Time.h * self.bh[i] * fi[i][:]
                    F[:] = ode_f90.vec_add(F[:],self.Time.h * self.bh[i],
                                           fi[i][:])
                if np.abs(self.b[i]) > 0.0:
                    b[:] = self.L * yi[i][:]
                    # include weak BCs here too.
                    F[:] = F[:] + self.Time.h * self.b[i] * b[:]
            yi[self.stages-1][:] = ystart[:] + self.M_inv * F[:]
            time = self.Time.time + self.Time.h
            return self.apply_bc(yi[self.stages-1][:],time)
        else :
            time = self.Time.time + self.Time.h
            return self.apply_bc(yi[self.stages-1][:],time)
    
    def force_vec( self, y_vec, t ):
        """force function."""
        # 0) Output vector
        v = np.zeros( ( self.vector_size ), float )
        # 1) Vector to functions
        y_comp = self.vector_to_component(y_vec)
        
        if not self.force_function == None:
            w = self.force_function(y_comp,t)
            dof_1=0
            dof_2=0
            for i in range(self.number_of_eq):
                dof_1 = dof_2
                dof_2 = dof_2 + self.y0[i].Space.dof
                v[dof_1:dof_2] = w[i][:]
        return v
        
    def apply_bc(self,y_vec,t):
        '''
        Apply BCs.
        '''
        # 1) Vector to functions
        y_comp = self.vector_to_component(y_vec)
        # 2) Apply BCs
        y_comp = self.bound_cond(y_comp,t)
        # 3) Functions to vector
        y_vec = self.component_to_vector(self.y_help)
        return y_vec
        
    def component_to_vector(self,y_comp):
        y_vec = np.zeros( self.vector_size, float )
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + y_comp[i].Space.dof
            y_vec[dof_1:dof_2] = y_comp[i].glob()
        return y_vec
    
    def vector_to_component(self,y_vec):
        #
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + self.y0[i].Space.dof
            temp = self.y0[i].Space.mapping_q( y_vec[dof_1:dof_2] )
            self.y_help[i].basis_coeff = np.copy(temp)
        return self.y_help
                                    
    def mass_matrix(self):
        M_glob = PysparseMatrix(size=self.vector_size)
        M_glob_inv = PysparseMatrix(size=self.vector_size)
        dof_1 = 0
        for i in range( self.number_of_eq ):
            mass = sempy.operators.Mass( self.y0[i].Space )
            # Mass matrix
            M_local = mass.matrix
            val,irow,jcol = M_local.find()
            M_glob.addAt(val, irow+dof_1, jcol+dof_1)
            # Inverse of mass matrix
            M_inv = mass.matrix_inv
            val,irow,jcol = M_inv.find()
            M_glob_inv.addAt(val, irow+dof_1, jcol+dof_1)
            dof_1 = dof_1 + self.y0[i].Space.dof
        return M_glob,M_glob_inv
            
    def runge_kutta_method(self,prk_method):
        '''
        Method.
        '''
        if prk_method == None:
            [a,b,c,ah,bh,ch,stages]= dirk111()
        if prk_method == 'dirk111':
            [a,b,c,ah,bh,ch,stages]= dirk111()
        if prk_method == 'dirk222':
            [a,b,c,ah,bh,ch,stages]= dirk222()
        if prk_method == 'dirk122':
            [a,b,c,ah,bh,ch,stages]= dirk122()
        if prk_method == 'dirk252':
            [a,b,c,ah,bh,ch,stages]= dirk252()
        if prk_method == 'dirk443':
            [a,b,c,ah,bh,ch,stages]= dirk443()
        if prk_method == 'erk4':
            [a,b,c,ah,bh,ch,stages]= erk4()
        if prk_method == 'irk2':
            [a,b,c,ah,bh,ch,stages]= irk2()
        
        return [a, b, c, ah, bh, ch, stages]
    
    def output_vector(self,y_in,adopt_filename=False):
        y_out=[]
        for j in range(self.number_of_eq):
            if adopt_filename:
                y_out.append( sempy.Function(y_in[j].Space,
                                         basis_coeff=0.0,
                                         filename=y_in[j].filename) )
            else:
                y_out.append( sempy.Function(y_in[j].Space,
                                         basis_coeff=0.0) )
        for j in range(self.number_of_eq):
            y_out[j].basis_coeff=np.copy(y_in[j].basis_coeff)    
        return y_out
    
    def __action_linear_operator(self,y):
        '''
        Action of the linear operator.
        '''
        v = [ ]
        w = [ ]
        for i in range(self.number_of_eq):
            v.append( np.zeros( self.y0[i].Space.dof, float ) )
            w.append( np.zeros( self.y0[i].Space.dof, float ) )
            
        # Convert global vector to its constituents
        dof_1=0
        dof_2=0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + self.y0[i].Space.dof
            w[i] = y[dof_1:dof_2]
            
        # Action of the linear operator
        for i in range(self.number_of_eq):
            for j in range(self.number_of_eq):
                v[i][:] = v[i][:] + self.L_in[i][j]* w[j]
        # Create the vector to be returned:
        v_out = np.zeros( self.vector_size, float )
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + self.y0[i].Space.dof
            v_out[dof_1:dof_2] = v[i]
            
        return v_out
    
    def __preconditioner(self,L_pre):
        '''
        Precond.
        '''
        # Mass matrices
        M=[]
        for i in range( self.number_of_eq ):
            mass = sempy.operators.Mass( self.y0[i].Space ).matrix
            mass = self.__preconditioner_mask(mass,i)
            M.append(mass)
        # Combined
        P = PysparseMatrix(size=self.vector_size)

        # Assemble             
        dof_1 = 0
        for i in range(self.number_of_eq):
            dof_2 = 0
            for j in range(self.number_of_eq):
                # Block diagonal entries
                if j==i:
                    T = sempy.operators.MultipleOperators(
                                            [M[i],L_pre[i][j]], 
                                            [1.0,-self.Time.h],
                                            assemble='yes').matrix
                    T = self.__preconditioner_mask( T, j )
                    val,irow,jcol = T.find()
                    P.addAt(val, irow+dof_1, jcol+dof_2)
                # non-diagonal entries
                if not j==i:
                    T = sempy.operators.MultipleOperators(
                                           [L_pre[i][j]], 
                                           [-self.Time.h],
                                           assemble='yes').matrix
                    val,irow,jcol = T.find()
                    P.addAt(val, irow+dof_1, jcol+dof_2)
                dof_2 = dof_2 + self.y0[j].Space.dof
            dof_1 = dof_1 + self.y0[i].Space.dof
        # Convert to ll_mat
        Q = pysparse.spmatrix.ll_mat( self.vector_size, 
                                      self.vector_size )
                
        val,irow,jcol = P.find()
        Q.update_add_at(val,irow,jcol)
        del val,irow,jcol
        # Factorize
        #if self.library == 'superlu':
        print 'factorize'
        self.LU = direct.superlu.factorize( Q.to_csr(),
                                       drop_tol=0.5)#self.drop_tol )
        #if self.library == 'umfpack':
        #self.LU = umfpack.factorize( Q )
        #return P
        L_pre_inv = spl.LinearOperator( 
                                (self.vector_size,self.vector_size),
                                matvec = self.__action_precond, 
                                dtype = 'float' )
        return L_pre_inv
    
    def __action_precond(self,v):
        w = np.zeros(self.vector_size, float)
        self.LU.solve(v,w)
        return w
    
    def __preconditioner_mask(self,Q,j):
        # Mask
        if self.y0[0].Space.dim == 1:
            for k in range(self.y0[j].Space.noe_bc):
                [i, phys, bc_type, element_edge] = \
                                  self.y0[j].Space.theta_phys_bc(k)
                if bc_type == "Dir":
                     Q[int(i),int(i)] = 1.0
                     
        if self.y0[0].Space.dim == 2:
            for k in range(self.y0[j].Space.noe_bc):
                for m in range(self.y0[j].Space.n):
                    [i, phys, bc_type, element_edge] = \
                                  self.y0[j].Space.theta_phys_bc(k,m)
                    if bc_type == "Dir":
                        Q[int(i),int(i)] = 1.0
        
        if self.y0[0].Space.dim == 3:
            for k in range(self.y0[j].Space.noe_bc):
                for m in range(self.y0[j].Space.n):
                    for n in range(self.y0[j].Space.n):
                        [i, phys, bc_type, element_edge] = \
                                  self.y0[j].Space.theta_phys_bc(k,m,n)
                    if bc_type == "Dir":
                        Q[int(i),int(i)] = 1.0
                        
        return Q
            
class PRK_old_1():
    '''
    Implementation of partitioned Runge-Kutta (PRK) method
    for the solution of ODE systems on the form:  
    
    .. math::
    
       M\\frac{dy}{dt}=Ly+N(y,t)
    
    :param y: Initial condition
    :type y: A collection of :class:`sempy.Function`
    :param Time: An instance of the time class.
    :type Time: :class:`sempy.ode.Time`
    :kwargs: * **L** - Linear operator
             * **L_pre** - The FEM equivalent of L. Used to create 
               preconditioner. 
             * **force_function** - A function, possibly non-linear. 
               This is the function :math:`N(y,t)` in the equation above. 
               Called as:: 
               
                 F=force_function(y,t)
                 
             * **bound_cond** - Strong BCs. Called as:: 
               
                 y=bound_cond(y,t)
                 
             * **prk_method** (*string*) - The PRK method to be used. 
               The options are:
                    * :literal:`DIRK111`
                    * :literal:`DIRK122`
                    * :literal:`DIRK222`
                    * :literal:`DIRK443`
             * **linear_solver** - The linear solver. The default is 
               CG. If L is not SPD, then one can for example use bicgstab.
             * **iter_tol** (*float*) - Iterative tolerance.
             * **maxiter** (*int*) - Maximum number of iterations.
    '''
    def __init__( self, y, Time, L = None, L_pre= None,
                  force_function = None, bound_cond = __bc__,
                  prk_method = None, 
                  linear_solver = None, iter_tol = 1e-10,
                  maxiter = 1000):
        # Solution vector
        self.number_of_eq = len(y)
        self.y0 = self.output_vector(y)
        self.y_out = self.output_vector(y)
        dof = 0
        for i in range(self.number_of_eq):
            dof = dof + self.y0[i].Space.dof 
        self.vector_size = dof
        # Strong BC
        self.bound_cond = bound_cond
        # Temporal
        self.Time = Time
        # PRK method
        self.prk_method = prk_method
        if self.prk_method == None:
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= dirk111()
        if self.prk_method == 'dirk111':
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= dirk111()
        if self.prk_method == 'dirk222':
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= dirk222()
        if self.prk_method == 'dirk122':
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= dirk122()
        if self.prk_method == 'dirk252':
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= dirk252()
        if self.prk_method == 'dirk443':
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= dirk443()
        if self.prk_method == 'erk4':
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= erk4()
        if self.prk_method == 'irk2':
            [self.a,self.b,self.c,
             self.ah,self.bh,self.ch,self.stages]= irk2()
        # Linear operator
        self.L_in = L
        if self.L_in == None:
            self.L_in = []
            G=[]
            for i in range( self.number_of_eq ):
                G.append(0)
            for i in range( self.number_of_eq ):
                self.L_in.append(G) 
            del G
        self.L = spl.LinearOperator( (self.vector_size,
                                      self.vector_size),
                                      matvec = self.__action_linear_operator,
                                      dtype = 'float')   
        # Mass matrix
        self.M = PysparseMatrix(size=self.vector_size)
        self.M_inv = PysparseMatrix(size=self.vector_size)
        dof_1 = 0
        for i in range( self.number_of_eq ):
            mass = sempy.operators.Mass( self.y0[i].Space )
            # Mass
            M = mass.matrix
            val,irow,jcol = M.find()
            self.M.addAt(val, irow+dof_1, jcol+dof_1)
            # Mass inverse
            M_inv = mass.matrix_inv
            val,irow,jcol = M_inv.find()
            self.M_inv.addAt(val, irow+dof_1, jcol+dof_1)
            dof_1 = dof_1 + self.y0[i].Space.dof
        # Preconditioner
        self.L_pre = L_pre
        if self.L_pre == None:
            self.L_pre_inv = 'none'
        else:
            self.L_pre_inv = self.__preconditioner(self.L_pre)
        # Non-linear function
        self.force_function = force_function
        # Linear solver:    
        self.linear_solver = linear_solver
        self.iter_tol = iter_tol
        self.maxiter=maxiter
        if self.linear_solver == None:
            self.linear_solver = sempy.linsolvers.Krylov(
                                                tol = self.iter_tol,
                                                solver_type = 'cg',
                                                maxiter=self.maxiter,
                                                pre=self.L_pre_inv)
    def output_vector(self,y_in):
        self.y_out=[]
        for j in range(self.number_of_eq):
            self.y_out.append( sempy.Function(y_in[j].Space,
                                              basis_coeff=0.0) )
        for j in range(self.number_of_eq):
            self.y_out[j].basis_coeff=np.copy(y_in[j].basis_coeff)    
        return self.y_out
            
    def __preconditioner_mask(self,Q,j):
        # Mask
        if self.y0[0].Space.dim == 1:
            for k in range(self.y0[j].Space.noe_bc):
                [i, phys, bc_type, element_edge] = \
                                  self.y0[j].Space.theta_phys_bc(k)
                if bc_type == "Dir":
                     Q[int(i),int(i)] = 1.0
                     
        if self.y0[0].Space.dim == 2:
            for k in range(self.y0[j].Space.noe_bc):
                for m in range(self.y0[j].Space.n):
                    [i, phys, bc_type, element_edge] = \
                                  self.y0[j].Space.theta_phys_bc(k,m)
                    if bc_type == "Dir":
                        Q[int(i),int(i)] = 1.0
        return Q
    
                                        
    def __preconditioner(self,L_pre):
        '''
        Precond.
        '''
        # Mass matrices
        M=[]
        for i in range( self.number_of_eq ):
            mass = sempy.operators.Mass( self.y0[i].Space ).matrix
            mass = self.__preconditioner_mask(mass,i)
            M.append(mass)
        # Combined
        P = PysparseMatrix(size=self.vector_size)

        # Assemble             
        dof_1 = 0
        for i in range(self.number_of_eq):
            dof_2 = 0
            for j in range(self.number_of_eq):
                # Block diagonal entries
                if j==i:
                    T = sempy.operators.MultipleOperators(
                                            [M[i],L_pre[i][j]], 
                                            [1.0,-self.Time.h],
                                            assemble='yes').matrix
                    T = self.__preconditioner_mask( T, j )
                    val,irow,jcol = T.find()
                    P.addAt(val, irow+dof_1, jcol+dof_2)
                # non-diagonal entries
                if not j==i:
                    T = sempy.operators.MultipleOperators(
                                           [L_pre[i][j]], 
                                           [-self.Time.h],
                                           assemble='yes').matrix
                    val,irow,jcol = T.find()
                    P.addAt(val, irow+dof_1, jcol+dof_2)
                dof_2 = dof_2 + self.y0[j].Space.dof
            dof_1 = dof_1 + self.y0[i].Space.dof
        # Convert to ll_mat
        Q = pysparse.spmatrix.ll_mat( self.vector_size, 
                                      self.vector_size )
                
        val,irow,jcol = P.find()
        Q.update_add_at(val,irow,jcol)
        del val,irow,jcol
        # Factorize
        #if self.library == 'superlu':
        print 'factorize'
        self.LU = direct.superlu.factorize( Q.to_csr(),
                                       drop_tol=0.5)#self.drop_tol )
        #if self.library == 'umfpack':
        #self.LU = umfpack.factorize( Q )
        #return P
        L_pre_inv = spl.LinearOperator( 
                                (self.vector_size,self.vector_size),
                                matvec = self.__action_precond, 
                                dtype = 'float' )
        return L_pre_inv
    
    def __action_precond(self,v):
        w = np.zeros(self.vector_size, float)
        self.LU.solve(v,w)
        return w
                                    
                
    def component_to_vector(self,y_comp):
        y_vec = np.zeros( self.vector_size, float )
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + y_comp[i].Space.dof
            y_vec[dof_1:dof_2] = y_comp[i].glob()
        return y_vec
    
    def vector_to_component(self,y_vec):
        #
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + self.y_out[i].Space.dof
            temp = self.y0[i].Space.mapping_q( y_vec[dof_1:dof_2] )
            self.y_out[i].basis_coeff = np.copy(temp)
        
    
    def __action_linear_operator(self,y):
        # Action of the linear operator.
        v = [ ]
        w = [ ]
        for i in range(self.number_of_eq):
            v.append( np.zeros( self.y0[i].Space.dof, float ) )
            w.append( np.zeros( self.y0[i].Space.dof, float ) )
        # Convert global vector to its constituents
        dof_1=0
        dof_2=0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + self.y0[i].Space.dof
            w[i] = y[dof_1:dof_2]
        # Action of the linear operator
        for i in range(self.number_of_eq):
            for j in range(self.number_of_eq):
                #if not self.L_in[i][j] == None:
                v[i][:] = v[i][:] + self.L_in[i][j]* w[j]
        # Create the vector to be returned:
        v_out = np.zeros( self.vector_size, float )
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + self.y0[i].Space.dof
            v_out[dof_1:dof_2] = v[i]
            
        return v_out
             
    def force_vec( self, y, t ):
        """force function."""
        v = np.zeros( ( self.vector_size ), float )
        if not self.force_function == None:
            w = self.force_function(y,t)
            dof_1=0
            dof_2=0
            for i in range(self.number_of_eq):
                dof_1 = dof_2
                dof_2 = dof_2 + self.y_out[i].Space.dof
                v[dof_1:dof_2] = w[i]
        return v
    
    #def __bc( self, y, t ):
    #    """This is an empty function."""
    #    # 1) Vector to constituents
    #    self.vector_to_component( y )# gives self.y_out
    #    self.y_out = self.bound_cond( self.y_out, t )
    #    # w is [sempy Functions]
    #    y = self.component_to_vector( self.y_out )
    #    return y
        
    def solve(self):
        '''
        Solves the ODE system. 
        '''
        #-- Solution vector y: y[0]=y^{n+1}, y[1]=y^{n}, y[2]=y^{n-1}, 
        # and so on
        ystart = np.zeros( (self.vector_size), float )
        yend   = np.zeros( (self.vector_size), float )
        ystart[:] = self.component_to_vector( self.y0 )
        # --Checking for b_i=a_si
        alpha=0.0
        for i in range(self.stages):
            alpha = alpha + np.abs(self.b[i]-self.a[self.stages-1,i]) + \
                            np.abs(self.bh[i]-self.ah[self.stages-1,i])
        #-- Time-stepping
        for i in range(self.Time.time_steps):
            print '--> time step no=',i+1
            yend = self.__timestepping__(ystart,alpha)
            self.Time.increment()
            ystart = yend
            self.vector_to_component(yend)
            #self.y0.to_file()
            print 'time=',self.Time.time
            #self.y_out[0].plot_wire()
            
        return self.y_out


    def __timestepping__(self,ystart,alpha):
        '''
        Advancing the solution one time-step.
        '''
        #-- Solution vector yi
        yi = np.zeros((self.stages,self.vector_size),float)
        fi = np.zeros((self.stages,self.vector_size),float)
        F  = np.zeros((self.vector_size),float)
        b  = np.zeros((self.vector_size),float)
        y_temp  = np.zeros((self.vector_size),float)
        for i in range(self.stages):
            yi[i,:] = np.copy(ystart[:])
            #yi[i,:] = ystart[:]
        #-- Computing the y_i's
        for i in range(self.stages):
            F[:] = 0.0
            for j in range(i):
                # Explicit
                if np.abs(self.ah[i,j]) > 0.0:
                    F[:] = F[:] + self.Time.h * self.ah[i,j] * fi[j][:]
                # Implicit
                if np.abs(self.a[i,j]) > 0.0:
                    b[:] = self.L*yi[j][:]
                    F[:] = F[:] + self.Time.h*self.a[i,j]*b[:]
                    # include weak BCs in this step. 
                    
            b[:] = self.M*ystart[:] + F[:]
            if np.abs(self.a[i,i]) > 0.0:
                # System solve required. Solve the system Hy[i]=b
                local_time  = self.Time.time + self.c[i] * self.Time.h
                yi[i][:] = self.bound_cond(yi[i][:],local_time)
                H = sempy.operators.MultipleOperators([self.M,self.L],\
                       scaling_factor=[1.0, -self.Time.h*self.a[i,i]],\
                       assemble='no').matrix
                [yi[i][:],flag] = self.linear_solver.solve(H,b,yi[i][:])
                #yi[i][:]=np.copy(y_temp[:])
                print 'iter=',flag
                # 1) If pressure-correction: include poisson solver and
                # update equations
                # 2) Include optional hyperbolic filter here. 
            else:
                # System solve not required
                time  = self.Time.time + self.ch[i] * self.Time.h
                yi[i][:] = self.M_inv*b[:]
                yi[i][:] = self.bound_cond(yi[i][:],time)
            time  = self.Time.time + self.ch[i] * self.Time.h
            fi[i][:] = self.force_vec(yi[i][:],time)
        # Computing the final value yend:
        if np.abs(alpha) > 0.0 :
            print 'last step required'
            F[:] = 0.0
            for i in range(self.stages):
                if np.abs(self.bh[i]) > 0.0:
                    F[:] = F[:] + self.Time.h * self.bh[i] * fi[i][:]
                if np.abs(self.b[i]) > 0.0:
                    b[:] = self.L * yi[i][:]
                    # include weak BCs here too.
                    F[:] = F[:] + self.Time.h * self.b[i] * b[:]
            yi[self.stages-1][:] = ystart[:] + self.M_inv * F[:]
            time = self.Time.time + self.Time.h
            return self.bound_cond(yi[self.stages-1][:],time)
        else :
            time = self.Time.time + self.Time.h
            return self.bound_cond(yi[self.stages-1][:],time)#yi[self.stages-1]


   

class PRK_old_2():
    ''' 
    Implementation of partitioned Runge-Kutta (PRK) methods
    for the solution of ODE systems.  
    
    Solves
    
    .. math:: 
       \underline{M}\\frac{d\underline{y}}{dt}=
       \underline{L}\,\underline{y}+\underline{N}(\underline{y},t) 
       
    where :math:`\underline{M}` is the SEM mass matrix, 
    :math:`\underline{L}` is a linear operator and 
    :math:`\underline{N}` is a non-linear function.
    
    :param y: Starting value.
    :type y: :class:`sempy.Function`
    :param Mass: An instance of the Mass class. This instance supplies the 
                 mass matrix and its inverse.   
    :type Mass: :class:`sempy.operators.Mass`
    :param matrix L: Linear operator.
    :param matrix L_pre: A matrix that approximates :literal:`L`. 
                         Used for preconditioning. 
    :param function non_lin: Function to be treated explicitly. Called as 
                     :literal:`non_lin(y,t)`.
    
    :param function bound_cond: Function with (possibly) time dependent boundary 
                       conditions. Called as :literal:`bound_cond(y,t)`. 
    :param string prk_method: The PRK method can be 
                              :literal:`dirk111`, 
                              :literal:`dirk122`, 
                              :literal:`dirk222`.
    :param LinearSolver: The linear solver for system solves at 
                         each time-step. It is called as 
                         :literal:`LinearSolver.solve(A,b,u)`. The default 
                         is CG.  
    :type LinearSolver: :class:`sempy.linsolvers.Krylov`, :class:`sempy.linsolvers.Direct`
    
    '''
    def __init__( self, y, Mass , Time, L = 'none', L_pre = 'none',
                  non_lin = 'none',
                  bound_cond = 'none', prk_method = 'dirk111',
                  LinearSolver = 'none' ):
        
        self.y0 = y
        self.Space = self.y0.Space
        self.dof = self.Space.dof
        self.Time = Time
        
        #start_time = 0.0, end_time = 1.0,time_steps = 5,
        #self.start_time = float(start_time)
        #self.end_time   = float(end_time)
        #self.time_steps = time_steps
        #self.h = (self.end_time-self.start_time)/float(self.time_steps)
        
        
        
        if non_lin == 'none':
            self.exp_func =  __emptyfunc__
        else:
            self.exp_func = non_lin
        
        #if Mass == 'none':
        #    self.M     = np.ones((self.dof),float)
        #    self.M_inv = np.ones((self.dof),float)
        #    self.M     = self.L.Space.mask(self.M)
        #    self.M_inv = self.L.Space.mask(self.M_inv)
        #else:
        self.M = Mass.matrix
        self.M_inv = Mass.matrix_inv
        self.L = L
        
        
        if bound_cond == 'none':
            self.bound_cond = __emptyfunc__
        else:
            self.bound_cond = bound_cond
            
        self.prk_method = prk_method
        
        if LinearSolver == 'none':
            self.LinearSolver = sempy.linsolvers.Krylov()
        else:
            self.LinearSolver = LinearSolver
        
        #self.L_pre = L_pre
           
        try:
            if L_pre == 'none':
                print 'L_pre'
                self.L_pre = L_pre
        except:
            #pass
            self.L_pre = L_pre
            self.LinearSolver.pre = sempy.precond.Preconditioner(
                                self.y0.Space, 
                                [self.M,self.L_pre],
                                scaling_factor=[1.0, -self.Time.h]).matrix
                    
                    
                    
            #self.LinearSolver.pre = sempy.precond.Preconditioner(
            #                    self.y0.Space, 
            #                    [self.M,self.L_pre],
            #                    scaling_factor=[1.0, -self.Time.h]).matrix
        #self.LinearSolver.pre = sempy.precond.Preconditioner([self.M,self.L.fem_matrix],\
        #                                    self.L.SpaceFEM,\
        #                                    scaling_factor=[1.0, -self.Time.h ]).matrix
        
        if self.prk_method == 'dirk111':
            [self.a,self.b,self.c,self.ah,self.bh,self.ch,self.stages]= \
                                                           dirk111()
        if self.prk_method == 'dirk122':
            [self.a,self.b,self.c,self.ah,self.bh,self.ch,self.stages]= \
                                                           dirk122()
        if self.prk_method == 'dirk222':
            [self.a,self.b,self.c,self.ah,self.bh,self.ch,self.stages]= \
                                                           dirk222()
        if self.prk_method == 'dirk252':
            [self.a,self.b,self.c,self.ah,self.bh,self.ch,self.stages]= \
                                                           dirk252()
   
    def solve(self):
        '''
        Solves the ODE system. 
        '''
        #-- Solution vector y: y[0]=y^{n+1}, y[1]=y^{n}, y[2]=y^{n-1}, 
        # and so on
        ystart = np.zeros( (self.dof), float )
        yend   = np.zeros( (self.dof), float )
        #ystart[:] = self.y0[:]
        ystart[:] = self.y0.glob()
        
        # --Checking for b_i=a_si
        alpha=0.0
        for i in range(self.stages):
            alpha = alpha + np.abs(self.b[i]-self.a[self.stages-1,i]) + \
                            np.abs(self.bh[i]-self.ah[self.stages-1,i])
        #-- Output
        self.y0.to_file()
        #-- Time-stepping
        for i in range(self.Time.time_steps):
            print '--> time step no=',i+1
            yend = self.__timestepping__(ystart,alpha)
            self.Time.increment()
            ystart = yend
            self.y0.basis_coeff = self.Space.mapping_q(ystart)
            self.y0.to_file()
            print 'time=',self.Time.time
            
        return self.y0


    def __timestepping__(self,ystart,alpha):
        #-- Solution vector yi
        yi = np.zeros((self.stages,self.dof),float)
        fi = np.zeros((self.stages,self.dof),float)
        F  = np.zeros((self.dof),float)
        b  = np.zeros((self.dof),float)
        for i in range(self.stages):
            yi[i,:] = ystart[:]
        #-- Advancing the solution one time step:
        #-- Computing the y_i's
        for i in range(self.stages):
            F[:] = 0.0
            for j in range(i):
                if np.abs(self.ah[i,j]) > 0.0:
                    F = F + self.Time.h * self.ah[i,j] * fi[j]
                #if self.L != 'none' and np.abs(self.a[i,j]) > 0.0:
                if np.abs(self.a[i,j]) > 0.0:
                    b = self.L*yi[j]
                    F = F + self.Time.h*self.a[i,j]*b
                    
            b = self.M*ystart + F
            #if self.L != 'none' and np.abs(self.a[i,i]) > 0.0:
            if np.abs(self.a[i,i]) > 0.0:
                # System solve required. Solve the system Hy[i]=b
                time  = self.Time.time + self.c[i] * self.Time.h
                yi[i] = self.bound_cond(yi[i],time)
                #H  = Helmholtz(self.M,self.L,self.Time.h,1.0,\
                #               self.a[i,i]).matrix
                H = sempy.operators.MultipleOperators([self.M,self.L],\
                       scaling_factor=[1.0, -self.Time.h*self.a[i,i]],\
                       assemble='no').matrix
                # If uzawa: solve uzawa
                #p=uzawa(H,b,G,D)
                [yi[i],flag] = self.linear_solver.solve(H,b,yi[i])
                print 'iter=',flag
            else:
                # System solve not required
                time  = self.Time.time + self.ch[i] * self.Time.h
                #yi[i] = self.bound_cond(b,time)
                yi[i] = self.M_inv*b
                yi[i] = self.bound_cond(yi[i],time)
            time  = self.Time.time + self.ch[i] * self.Time.h
            fi[i] = self.exp_func(yi[i],time)
        #--- Computing the final value yend:
        if alpha > 0 :
            # print 'last step required'
            F[:] = 0.0
            for i in range(self.stages):
                if np.abs(self.bh[i]) > 0.0:
                    F = F + self.Time.h * self.bh[i] * fi[i]
                #if self.L != 'none' and np.abs(self.b[i]) > 0.0:
                if np.abs(self.b[i]) > 0.0:
                    b = self.L * yi[i]
                    F = F + self.Time.h * self.b[i] * b
            yi[self.stages-1] = ystart + self.M_inv * F
            time = self.Time.time + self.Time.h
            return self.bound_cond(yi[self.stages-1], time)
        else :
            return yi[self.stages-1]


    




'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 

import scipy.sparse.linalg as spl
import scipy.sparse as sps
import numpy as np
import operators_f90
# Pysparse
import pysparse
# Compatibility with older versions
if int(pysparse.__version__[2]) >=2:
    from pysparse.sparse.pysparseMatrix import PysparseMatrix #as pysparseMatrix
    
if int(pysparse.__version__[2]) <=1:
    from pysparse.pysparseMatrix import PysparseMatrix
#from pysparse import pysparseMatrix
import time

import sempy

#import PyTrilinos.Epetra as Epetra


class Laplacian():
    '''
    The Laplace operator. 
       
    :param Space: Discrete space.
    :type Space: :class:`sempy.Space`
    :param mu: Transport coefficient, such as viscosity or thermal conductivity. 
    :type mu: :class:`sempy.Function`
    :param g_neu: Neumann boundary conditions.
    :type g_neu: :class:`sempy.BoundaryFunction`
    :param g_rob: Robin boundary conditions.
    :type g_rob: :class:`sempy.BoundaryFunction`
    :param u_infty: Robin boundary conditions.
    :type u_infty: :class:`sempy.BoundaryFunction`
    :param mask: If :literal:`mask='yes'` the matrix is masked according to 
                 strong boundary conditions. If :literal:`mask='no'` the
                 matrix is not masked (results in Neumann-Laplace matrix 
                 operator).  The mask flag has only effect on the matrix free 
                 method and the linear_form. 
    :type mask: string
    :param string assemble: If :literal:`assemble='no'` (default) the method 
                            is matrix free. For :literal:`assemble = 'yes'` 
                            the stiffness matrix is assembled and stored in 
                            the CSR format. The latter might be used with care 
                            for high values of :attr:`sempy.Space.n` 
                            and :attr:`sempy.Space.noe` as the SEM Laplacian 
                            is elementwise full. 
    :param string fem_assemble: If :literal:`assemble = 'yes'` the 
                                corresponding FEM matrix is assembled.  
    
    :attributes: * **matrix** (LinearOperator or PysparseMatrix ) - Returns 
                   the stiffness matrix  
                   :math:`A\in \\mathbb R^{\\text{dof}\\times\\text{dof}}`. 
                   Wether *matrix* holds any data at all depends on the 
                   :literal:`assemble` paramter above.
                 * **linear_form** (array) - Returns a one dimensional array 
                   of length :attr:`sempy.Space.dof`  
                   (global degrees of freedom).
                 * **matrix_fem** (LinearOperator or PysparseMatrix ) - The 
                   finite element matrix based on the same nodal points 
                   as the GLL mesh.

    This class supplies the stiffness :literal:`matrix` and 
    the :literal:`linear_form` resulting from applying SEM to the Laplace 
    operator. 
    
    .. math::
       -\\int_{\\Omega}(\\nabla\\cdot\\mu\\nabla u)v d\\Omega =
       \\int_{\\Omega} \\mu\\nabla u\\cdot \\nabla v d\\Omega -
       \\int_{\\Gamma} (\\mu\\nabla u\cdot\mathbf n )vd\\Gamma
       
    Three different boundary conditions are associated with the Laplacian, 
    the Dirichlet, the Neumann and the Robin boundary conditions, i.e.:
    
    .. math::
       u=u_\\text{dir}\quad\\text{(Dirichlet)}
       
       \\mu(\\partial u/\\partial n)|_{\\Gamma}=g_\\text{neu}
       \quad\\text{(Neumann)}
       
       \\mu(\\partial u/\\partial n)|_{\\Gamma}=g_\\text{rob}(u_{\\infty}-u)
       \quad\\text{(Robin)}
    
    which, when inserted into the above equation gives   
    
    .. math::
       -\\int_{\\Omega}(\\nabla\\cdot\\mu\\nabla u)v d\\Omega =
       \\underbrace{\\int_{\\Omega} \\mu\\nabla u\\cdot \\nabla v 
       +\\int_{\\Gamma_{rob}} g_\\text{rob}uvd\\Gamma}_{a(u,v) \\rightarrow Au}
       -\\underbrace{\\bigg(\\int_{\\Gamma_\\text{neu}} g_\\text{neu}vd\\Gamma 
       +\\int_{\\Gamma_\\text{rob}} g_\\text{rob}u_\\infty vd\\Gamma\\bigg)}_{l(v) \\rightarrow f}
       
      
    Dirichlet conditions are attached to the trial function (an instance of 
    the class :class:`sempy.Function`). Natural boundary conditions, on the 
    other hand, have to be supplied explicitly when creating an instance of 
    this class, i.e. :literal:`g_neu`, :literal:`g_rob`, 
    :literal:`u_infty`. 
    
    
                
    **Example**::
           
       import sempy
       
       X = sempy.Space( 'square.msh', n = 5, dim = 2 )
       mu = sempy.Function( X, basis_coeff = 1.0 )
       laplacian = sempy.operators.Laplacian( X, mu )
       A = laplacian.matrix
       f = laplacian.linear_form
    
    '''
    def __init__(self, Space, mu='none', g_neu='none', g_rob='none',\
                 u_infty='none', assemble='no' ,fem_assemble = 'no',
                 mask = 'yes',silent=False):
    #def __init__(self, Space, mu='none', g_neu='none', g_rob='none',\
    #             u_infty='none', assemble='no', SpaceFEM='none'):
        self.Space = Space
        self.assemble = assemble
        self.fem_assemble = fem_assemble
        #self.SpaceFEM = SpaceFEM
        self.mask = mask
        self.silent=silent
        
        if mu == 'none':
            self.mu = sempy.Function(self.Space, basis_coeff = 1.0)
        else:
            self.mu = mu
                
        if g_neu == 'none':
            self.g_neu = sempy.BoundaryFunction(self.Space)
        else:
            self.g_neu = g_neu
                
        if g_rob == 'none':
            self.g_rob = sempy.BoundaryFunction(self.Space)
        else:
            self.g_rob = g_rob
             
        if u_infty == 'none':
            self.u_infty = sempy.BoundaryFunction(self.Space)
        else:
            self.u_infty = u_infty
        
        if self.Space.dim == 1:
                       
            self.g11 = 1.0/self.Space.jac
            if self.assemble == 'no':
                self.matrix = spl.LinearOperator( (self.Space.dof,self.Space.dof),\
                                 matvec=self.__action_1d,dtype='float')    
            if self.assemble == 'yes':
                self.matrix = self.__assemble_1d()
                self.matrix.dtype='float64'
            self.linear_form = self.__rhs_1d()

        if self.Space.dim == 2:
                      
            self.g11, self.g12, \
            self.g21, self.g22   = \
                operators_f90.geometric_laplacian_2d(self.Space.x,self.Space.y,\
                                                self.Space.jac,self.Space.D)
            if self.assemble == 'no':
                self.matrix = spl.LinearOperator( (self.Space.dof,self.Space.dof),\
                                         matvec=self.__action_2d,dtype='float')    
            if self.assemble == 'yes':
                self.matrix = self.__assemble_2d()
                self.matrix.dtype='float64'
            self.linear_form = self.__rhs_2d()
            
        
        
        if self.Space.dim == 3:
            self.g11, self.g12, self.g13,\
            self.g21, self.g22, self.g23,\
            self.g31, self.g32, self.g33=\
                operators_f90.geometric_laplacian_3d(self.Space.x,self.Space.y,self.Space.z,\
                                                     self.Space.jac,self.Space.D)
            if self.assemble == 'no':
                self.matrix = spl.LinearOperator( (self.Space.dof,self.Space.dof),\
                                        matvec=self.__action_3d,dtype='float')  
            if self.assemble == 'yes':
                self.matrix = self.__assemble_3d()
                self.matrix.dtype='float64'
                
            self.linear_form = self.__rhs_3d()     
        
        # FEM matrix
        #if self.SpaceFEM == 'none':
        if self.fem_assemble == 'no':
            self.matrix_fem = 'none'
        #if not self.SpaceFEM == 'none':
        if self.fem_assemble == 'yes':
            #self.matrix_fem = 
            self.assemble_fem()
        
        # remove?
        self.matrix.Space = self.Space
        
            
    def assemble_fem(self):
        '''
        Assembles the FEM matrix used for preconditioning.
        
        Usage::
        
           import sempy 

           X = sempy.Space( filename = 'square', n = 10, dim = 2 )
           
           laplace = sempy.operators.Laplacian( X, assemble='yes',fem_assemble='yes')

           A = laplace.matrix
           A_fem = laplace.matrix_fem

           print 'Non-zero entries in A=', A.nnz
           print 'Non-zero entries in A_fem=',A_fem.nnz
            
        '''
        #basis_coeff = self.SpaceFEM.sem_to_fem( self.mu.basis_coeff )
        #mu_fem = sempy.Function( self.SpaceFEM, basis_coeff = basis_coeff )
        ##A_fem = Laplacian( self.SpaceFEM, mu = mu_fem, 
        ##                   assemble = 'yes').matrix
        #self.matrix_fem = Laplacian( self.SpaceFEM, mu = mu_fem, 
        #                   assemble = 'yes').matrix
        ##
        basis_coeff = self.Space.fem.sem_to_fem( self.mu.basis_coeff )
        mu_fem = sempy.Function( self.Space.fem, basis_coeff = basis_coeff )
        self.matrix_fem = Laplacian( self.Space.fem, mu = mu_fem, 
                           assemble = 'yes',
                           mask = self.mask).matrix                   
        #return A_fem
        
        
    def assemble_matrix(self):
        '''
        Assembles the stiffness matrix.
        
        Usage::
        
          import sempy 
      
          X = sempy.Space( filename = 'square_neumann', n = 3, dim = 2 )

          laplacian = sempy.operators.Laplacian( X )
          laplacian.assemble_matrix()
          A = laplacian.matrix 
        '''
        if self.Space.dim == 1:
                self.matrix = self.__assemble_1d()
                self.matrix.dtype='float64'
                
        if self.Space.dim == 2:
            self.matrix = self.__assemble_2d()
            self.matrix.dtype='float64'
            
        if self.Space.dim == 3:
                self.matrix = self.__assemble_3d()
                self.matrix.dtype='float64'
        return 
    
    def __action_1d(self,v_in):
        u = self.Space.mapping_q(v_in)
        # Local matrix vector product w_L=A_L*v_L
        w = operators_f90.action_of_laplacian_1d(u,self.mu.basis_coeff,\
                                                 self.Space.weights,\
                                                 self.g11,self.Space.D)
        # Mapping
        v_out = self.Space.mapping_qt(w)
        # Robin
        for k in range(self.Space.noe_bc):
            i = self.Space.theta_bc[k]
            v_out[i] = v_out[i] + self.g_rob.basis_coeff[k]*v_in[i]
        # Mask
        if self.mask == 'yes':
            v_out = self.Space.mask(v_out)
            
        return v_out
    
    def __action_2d(self,v_in):
        u = self.Space.mapping_q(v_in)
        # Local matrix vector product w_L=A_L*v_L
        w = operators_f90.action_of_laplacian_2d(u, self.mu.basis_coeff, \
                                                 self.Space.weights, self.Space.D,\
                                                 self.g11, self.g12, self.g21, self.g22)
        # Mapping
        v_out = self.Space.mapping_qt(w)
        # Robin
        v_out = operators_f90.action_of_laplacian_2d_robin(v_out,v_in,self.Space.theta_bc,\
                    self.Space.weights,self.Space.jac_s,self.g_rob.basis_coeff)
        # Mask
        if self.mask == 'yes':
            v_out = self.Space.mask(v_out)
            
        return v_out
    
    def __action_3d(self,v_in):
        u = self.Space.mapping_q(v_in)
        # Local matrix vector product w_L=A_L*v_L
        w = operators_f90.action_of_laplacian_3d(u,self.mu.basis_coeff,self.Space.weights,\
                                                 self.Space.D,\
                                                 self.g11,self.g12,self.g13,\
                                                 self.g21,self.g22,self.g23,\
                                                 self.g31,self.g32,self.g33)
        # Mapping
        v_out = self.Space.mapping_qt(w)
        # Robin
        v_out = operators_f90.action_of_laplacian_3d_robin(v_out,v_in,self.Space.theta_bc,\
                    self.Space.weights,self.Space.jac_s,self.g_rob.basis_coeff)
        # Mask
        if self.mask == 'yes':
            v_out = self.Space.mask(v_out)
            
        return v_out
        
        
    def __assemble_1d(self):
        '''
        assemble
        '''
        if not self.silent:
            print ''
            print '**** Assemble Laplacian       ****'
        nz = self.Space.noe * self.Space.n * self.Space.n
        
        #A = pysparse.spmatrix.ll_mat(self.Space.dof, self.Space.dof, nz)
        #A = sps.lil_matrix((self.Space.dof, self.Space.dof), dtype='float')
        #A = pysparseMatrix.PysparseMatrix(size=self.Space.dof)#,dtype=float)
        A = PysparseMatrix(size=self.Space.dof)#,dtype=float)
        #A = sps.lil_matrix(B)#,shape=(self.Space.dof, self.Space.dof), dtype='float')
        t1 = time.clock()
        irow,jcolumn,value = \
            operators_f90.assemble_laplacian_1d(self.Space.weights,self.mu.basis_coeff,\
                                                self.g11,self.Space.D,\
                                                self.Space.theta,nz)
        
        # Assign values
        for k in range(len(value)):
            i = int(irow[k])
            j = int(jcolumn[k])
            A[i,j] = A[i,j] + value[k]
        # Robin
        for k in range(self.Space.noe_bc):
            i = int(self.Space.theta_bc[k])
            A[i,i] = A[i,i] + self.g_rob.basis_coeff[k]
        # Mask
        A = self.Space.mask_matrix(A)
        t2 = time.clock()
        if not self.silent:
            print '**** ex. time=',t2-t1,'nnz=',A.nnz
            
        return A
        
    
        
    def __assemble_2d(self):
        
        if not self.silent:
            print ''
            print '**** Assemble Laplacian       ****'
        t1 = time.clock()
        
        # Local comnectivity array
        theta_local = np.zeros((self.Space.n, self.Space.n),int)
        count =0
        for m in range(self.Space.n):
            for n in range(self.Space.n):
                theta_local[m,n]=count
                count=count+1

        # Mask array
        
        mask   = np.zeros((self.Space.noe,self.Space.n,self.Space.n),int)
        if self.mask == 'yes':
            mask_g = np.zeros((self.Space.dof),int)
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    [i, phys, bc_type, edge] = self.Space.theta_phys_bc(k,m)
                    if bc_type == "Dir":
                        mask_g[i] = 1
                        
            for k in range(self.Space.noe):
                for m in range(self.Space.n):
                    for n in range(self.Space.n):
                        i = self.Space.theta[k,m,n]
                        mask[k,m,n] = mask_g[i]
            del mask_g
                
        # Matrix
        #A = pysparseMatrix.PysparseMatrix(size=self.Space.dof)#, sizeHint=nz)
        A = PysparseMatrix(size=self.Space.dof)#, sizeHint=nz)
                
        # Local degrees of freedom
        nz = self.Space.n*self.Space.n
        
        # Assemble stiffness matrix elementwize
        for k in range(self.Space.noe):
            val, row, col = operators_f90.assemble_laplacian_2d_new(
                                        self.Space.weights,
                                        self.mu.basis_coeff[k,:,:],
                                        self.g11[k,:,:], self.g12[k,:,:],
                                        self.g21[k,:,:], self.g22[k,:,:],
                                        self.Space.D,
                                        self.Space.theta[k,:,:],
                                        theta_local, mask[k,:,:], nz)
            val = np.reshape(val, nz*nz)
            row = np.reshape(row, nz*nz)
            col = np.reshape(col, nz*nz)
            A.addAt(val,row,col)
                
        # Add Robin contribution
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                [i, phys, bc_type, element_edge]=self.Space.theta_phys_bc(k,m)
                if element_edge[1] == 0:
                    I=self.Space.theta[element_edge[0],m,0]
                if element_edge[1] == 1:
                    I=self.Space.theta[element_edge[0],m,self.Space.n-1]
                if element_edge[1] == 2:
                    I=self.Space.theta[element_edge[0],0,m]
                if element_edge[1] == 3:
                    I=self.Space.theta[element_edge[0],self.Space.n-1,m]
                J=i
                A[int(I),int(J)] = A[int(I),int(J)] + \
                                   self.Space.weights[m]*self.Space.jac_s[k,m]*\
                                   self.g_rob.basis_coeff[k,m]   
        # Mask
        #print '**** nnz=',A.nnz
        #t1_m = time.clock()    
        #A = self.Space.mask_matrix(A)
        #t2_m = time.clock()    
        t2 = time.clock()    
        #print '**** assemble time=',t2_f-t1_f
        #print '**** robin    time=',t2_r-t1_r
        #print '**** mask     time=',t2_m-t1_m
        if not self.silent:
            print '**** assemble time=',t2-t1
            print '**** nnz=',A.nnz
        return A
    
    def __assemble_3d(self):
        
        if not self.silent:
            print ''
            print '**** Assemble Laplacian       ****'
        
        # Local comnectivity array
        theta_local = np.zeros((self.Space.n, self.Space.n, self.Space.n),int)
        count =0
        for k in range(self.Space.n):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    theta_local[k,m,n]=count
                    count=count+1
                    
        # Mask array
        mask   = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),int)
        mask_g = np.zeros((self.Space.dof),int)
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    [i, phys, bc_type, edge] = self.Space.theta_phys_bc(k,m,n)
                    if bc_type == "Dir":
                        mask_g[i] = 1
                        
        for k in range(self.Space.noe):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    for o in range(self.Space.n):
                        i = self.Space.theta[k,m,n,o]
                        mask[k,m,n,o] = mask_g[i]
        del mask_g                
        
        # Matrix
        #A = pysparseMatrix.PysparseMatrix(size=self.Space.dof)#, sizeHint=nz)
        A = PysparseMatrix(size=self.Space.dof)#, sizeHint=nz)
        
        # Local degrees of freedom
        nz =  self.Space.n*self.Space.n*self.Space.n
        t1 = time.clock()
        
        # Assemble the stiffness matrix, elementwize.
        t1_f = time.clock()    
        for k in range(self.Space.noe):
            val,row,col = \
                operators_f90.assemble_laplacian_3d(self.Space.weights,self.mu.basis_coeff[k,:,:,:],\
                                 self.g11[k,:,:,:],self.g12[k,:,:,:],self.g13[k,:,:,:],\
                                 self.g21[k,:,:,:],self.g22[k,:,:,:],self.g23[k,:,:,:],\
                                 self.g31[k,:,:,:],self.g32[k,:,:,:],self.g33[k,:,:,:],\
                                 self.Space.D,self.Space.theta[k,:,:,:],theta_local,mask[k,:,:,:],nz)
            val = np.reshape(val, nz*nz)
            row = np.reshape(row, nz*nz)
            col = np.reshape(col, nz*nz)
            A.addAt(val,row,col)
            
        t2_f = time.clock()    
        # Add Robin contribution
        t1_r = time.clock()    
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k,m,n)
                    if element_edge[1] == 0:
                        I = self.Space.theta[element_edge[0],m,n,0]
                    if element_edge[1] == 1:
                        I = self.Space.theta[element_edge[0],m,n,self.Space.n-1]
                    if element_edge[1] == 2:
                        I = self.Space.theta[element_edge[0],m,0,n]
                    if element_edge[1] == 3:
                        I = self.Space.theta[element_edge[0],m,self.Space.n-1,n]
                    if element_edge[1] == 4:
                        I = self.Space.theta[element_edge[0],0,m,n]
                    if element_edge[1] == 5:
                        I = self.Space.theta[element_edge[0],self.Space.n-1,m,n]
                    J = i
                    A[int(I),int(J)] = A[int(I),int(J)] + \
                                       self.Space.weights[m]*self.Space.weights[n]*\
                                       self.Space.jac_s[k,m,n]*\
                                       self.g_rob.basis_coeff[k,m,n]
        t2_r = time.clock()    
        # Mask
        if not self.silent:
            print '**** Number of nnz=',A.nnz
        #t1_m = time.clock()    
        #A = self.Space.mask_matrix(A)
        #t2_m = time.clock()    
        t2 = time.clock()
        #print '**** post mask nnz=',A.nnz
        if not self.silent:
            print '**** assemble time=',t2_f-t1_f
            print '**** robin    time=',t2_r-t1_r
            #print '**** mask     time=',t2_m-t1_m
            print '**** total    time=',t2-t1#,'nnz=',A.nnz
        
        return A
    
            
    def __rhs_1d(self):
        f = np.zeros((self.Space.dof), float)
        # Contributions from Neumann conditions
        for k in range(self.Space.noe_bc):
            i = self.Space.theta_bc[k]
            f[i] = f[i] + self.g_neu.basis_coeff[k]
        # Contributions from Robin conditions
        for k in range(self.Space.noe_bc):
            i = self.Space.theta_bc[k]
            f[i] = f[i] + self.g_rob.basis_coeff[k]*self.u_infty.basis_coeff[k]
        # Mask Dirichlet boundary points
        if self.mask == 'yes':
            f = self.Space.mask(f)
        return f
        
    def __rhs_2d(self):
        f = np.zeros((self.Space.dof), float)
        # Contributions from Neumann conditions
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                i = self.Space.theta_bc[k,m]
                f[i] = f[i] + self.Space.weights[m]*self.Space.jac_s[k,m]*\
                              self.g_neu.basis_coeff[k,m]
        # Contributions from Robin conditions
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                i = self.Space.theta_bc[k,m]
                f[i] = f[i] + self.Space.weights[m]*self.Space.jac_s[k,m]*\
                              self.g_rob.basis_coeff[k,m]*\
                              self.u_infty.basis_coeff[k,m]
        # Mask Dirichlet boundary points
        if self.mask == 'yes':
            f = self.Space.mask(f)
        return f
    
    def __rhs_3d(self):
        #f = np.zeros((self.Space.dof), float)
        # Contributions from Neumann conditions
        #for k in range(self.Space.noe_bc):
        #    for m in range(self.Space.n):
        #        for n in range(self.Space.n):
        #            i = self.Space.theta_bc[k,m,n]
        #            f[i] = f[i] + self.Space.weights[m]*self.Space.weights[n]*\
        #                          self.Space.jac_s[k,m,n]*self.g_neu.basis_coeff[k,m,n]
        # Contributions from Robin conditions
        #for k in range(self.Space.noe_bc):
        #    for m in range(self.Space.n):
        #        for n in range(self.Space.n):
        #            i = self.Space.theta_bc[k,m,n]
        #            f[i] = f[i] + self.Space.weights[m]*self.Space.weights[n]*\
        #                          self.Space.jac_s[k,m,n]*\
        #                          self.g_rob.basis_coeff[k,m,n]*self.u_infty.basis_coeff[k,m,n]
        # 
        f = np.zeros(self.Space.dof, float)
        if self.Space.noe_bc > 0:
            f = operators_f90.assemble_linear_form_3d( \
                                self.Space.theta_bc,
                                self.Space.weights,\
                                self.Space.jac_s,\
                                self.g_rob.basis_coeff,\
                                self.u_infty.basis_coeff,\
                                self.g_neu.basis_coeff,self.Space.dof)
        # Mask Dirichlet boundary points
        if self.mask == 'yes':
            f = self.Space.mask(f)
        return f
        
    def assemble_linear_form(self):
        '''
        Assembles the linear form.
        '''
        
        if self.Space.dim == 1:
            f = self.__rhs_1d()
        if self.Space.dim == 2:
            f = self.__rhs_2d()
        if self.Space.dim == 3:
            f = self.__rhs_3d()
        return f
        
        
        
class Convection:
    '''
    The convection operator. 
    
    :param Space: The function space.
    :type Space: :class:`sempy.Space`
    :param u_conv: Convecting velocity field.
    :type u_conv: A list of :class:`sempy.Function` instances
    :param string assemble: A boolean argument controlling the assembly 
                            of the matrix. If :literal:`assemble = 'yes'`
                            the matrix is assembled and if 
                            :literal:`assemble = 'no'` the 
                            :literal:`matrix` only performs the action of 
                            the matrix. 
    :param string fem_assemble:  If :literal:`assemble = 'yes'` the 
                                 corresponding FEM matrix is assembled.  
             
    :attributes: * **matrix** - A matrix in the Pysparse format 
                   (:literal:`assemble = 'yes'`) or an instance of the class
                   :class:`scipy.sparse.linalg.LinearOperator` 
                   (:literal:`assemble = 'no'`). 
                 * **matrix_fem** - The corresponding FEM matrix.
    
    .. :todo: Option to controll conservative/non-conservative form.
    
    **Example**::
    
        import sempy
        
        X = sempy.Space( 'square', n = 10, dim = 2 )
        
        # Convecting velocity field
        u = sempy.Function( X, basis_coeff = 1.0 )
        v = sempy.Function( X, basis_coeff = 1.0 )
        U = [ u, v ]
        
        conv = sempy.operators.Convection( X, u_conv = U,
                                           assemble = 'yes',
                                           fem_assemble = 'yes' )
        # Convection matrix
        C = conv.matrix
        C_fem = conv.matrix_fem
    
    '''
    def __init__(self,Space,u_conv='none',assemble='no',
                 fem_assemble = 'no', silent = False ):
        self.Space = Space
        self.dim = self.Space.dim
        self.silent = silent
        
        if u_conv == 'none':
            if self.Space.dim == 1:
                _u = sempy.Function(self.Space, basis_coeff = 1.0)
                self.u_conv = [ _u.basis_coeff ]
            if self.Space.dim == 2:
                _u = sempy.Function(self.Space, basis_coeff = 1.0)
                _v = sempy.Function(self.Space, basis_coeff = 1.0)
                self.u_conv = [ _u.basis_coeff, _v.basis_coeff ]
            if self.Space.dim == 3:
                _u = sempy.Function(self.Space, basis_coeff = 1.0)
                _v = sempy.Function(self.Space, basis_coeff = 1.0)
                _w = sempy.Function(self.Space, basis_coeff = 1.0)
                self.u_conv = [ _u.basis_coeff, _v.basis_coeff, _w.basis_coeff ]
        else:
            if self.Space.dim == 1:
                self.u_conv=[ u_conv[0].basis_coeff ]
            if self.Space.dim == 2:
                self.u_conv=[ u_conv[0].basis_coeff,
                          u_conv[1].basis_coeff ]
            if self.Space.dim == 3:
                self.u_conv=[ u_conv[0].basis_coeff,
                          u_conv[1].basis_coeff,
                          u_conv[2].basis_coeff ]
             
        self.assemble = assemble
        self.fem_assemble = fem_assemble
        #self.SpaceFEM = self.Space.fem
        
        if self.dim == 1:
            if self.assemble == 'no':
                self.matrix = spl.LinearOperator( (self.Space.dof,self.Space.dof),\
                                    matvec=self.__action_1d,dtype='float')  
            if self.assemble == 'yes':
                self.matrix = self.__assemble_1d()
            
        
        if self.dim == 2:
            self.g11,self.g12,self.g21,self.g22=\
               operators_f90.geometric_convection_2d(self.Space.x,
                                                     self.Space.y,
                                                     self.Space.D)
            if self.assemble == 'no':
                self.matrix = spl.LinearOperator( (self.Space.dof,self.Space.dof),\
                                               matvec=self.__action_2d,
                                                   dtype='float')  
            if self.assemble == 'yes':
                self.matrix = self.__assemble_2d()
        
        if self.dim == 3:
            self.g11 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g12 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g13 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g21 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g22 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g23 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g31 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g32 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            self.g33 = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),
                                float)
            
            for k in range(self.Space.noe):
                self.g11[k,:,:,:],self.g12[k,:,:,:],self.g13[k,:,:,:],\
                self.g21[k,:,:,:],self.g22[k,:,:,:],self.g23[k,:,:,:],\
                self.g31[k,:,:,:],self.g32[k,:,:,:],self.g33[k,:,:,:] =\
                    operators_f90.geometric_convection_3d(self.Space.x[k,:,:,:],
                                                          self.Space.y[k,:,:,:],
                                                          self.Space.z[k,:,:,:],
                                                          self.Space.D)
            # Weights array
            self.W=np.zeros((self.Space.n,self.Space.n,self.Space.n))
            for i in range(self.Space.n):
                for j in range(self.Space.n):
                    for k in range(self.Space.n):
                        self.W[i,j,k]=self.Space.weights[i]*self.Space.weights[j]*self.Space.weights[k]
                    
            if self.assemble == 'no':
                self.matrix = spl.LinearOperator( (self.Space.dof,self.Space.dof),
                                    matvec=self.__action_3d,dtype='float')  
            if self.assemble == 'yes':
                self.matrix = self.__assemble_3d()
        
        #if not self.SpaceFEM == 'none':
        if not self.fem_assemble == 'no':
            #self.matrix_fem = self.assemble_fem()
            self.assemble_fem()
        else:
            self.matrix_fem = 'none'
    
    def assemble_matrix(self):
        '''
        Assembles the stiffness matrix.
        
        Usage::
        
          import sempy 
      
          X = sempy.Space( filename = 'square', n = 3, dim = 2 )

          conv = sempy.operators.Convection( X )
          conv.assemble_matrix()
          C = conv.matrix 
        '''
        if self.Space.dim == 1:
                self.matrix = self.__assemble_1d()
                self.matrix.dtype='float64'
                
        if self.Space.dim == 2:
            self.matrix = self.__assemble_2d()
            self.matrix.dtype='float64'
            
        if self.Space.dim == 3:
                self.matrix = self.__assemble_3d()
                self.matrix.dtype='float64'
        return 
            
    def assemble_fem(self):
        '''
        Assembles the FEM matrix used for preconditioning.
        
        Usage::
        
           import sempy

           X = sempy.Space( 'square', n = 7, dim = 2 )
           conv = sempy.operators.Convection(X)
           conv.assemble_fem()
           
           C_fem = conv.matrix_fem
        '''
        if self.dim == 1:
            #u_conv[0] = self.SpaceFEM.sem_to_fem(self.u_conv[0])
            
            U = sempy.Function( self.Space.fem, 
                 basis_coeff = self.Space.fem.sem_to_fem(self.u_conv[0]) )
            u_conv = [ U ]
            
        if self.dim == 2:
            #u_conv = np.ones((2, self.SpaceFEM.noe, self.SpaceFEM.n, self.SpaceFEM.n),float)
            #u_conv[0] = self.SpaceFEM.sem_to_fem(self.u_conv[0])
            #u_conv[1] = self.SpaceFEM.sem_to_fem(self.u_conv[1])
            U = sempy.Function( self.Space.fem, 
                 basis_coeff = self.Space.fem.sem_to_fem(self.u_conv[0]) )
            V = sempy.Function( self.Space.fem, 
                 basis_coeff = self.Space.fem.sem_to_fem(self.u_conv[1]) )
            u_conv = [ U, V ]
            
        if self.dim == 3:
            #u_conv = np.ones((3, self.SpaceFEM.noe, \
            #                     self.SpaceFEM.n, self.SpaceFEM.n, self.SpaceFEM.n),float)
            #u_conv[0] = self.SpaceFEM.sem_to_fem(self.u_conv[0])
            #u_conv[1] = self.SpaceFEM.sem_to_fem(self.u_conv[1])
            #u_conv[2] = self.SpaceFEM.sem_to_fem(self.u_conv[2])
            U = sempy.Function( self.Space.fem, 
                 basis_coeff = self.Space.fem.sem_to_fem(self.u_conv[0]) )
            V = sempy.Function( self.Space.fem, 
                 basis_coeff = self.Space.fem.sem_to_fem(self.u_conv[1]) )
            W = sempy.Function( self.Space.fem, 
                 basis_coeff = self.Space.fem.sem_to_fem(self.u_conv[2]) )
            u_conv = [ U, V, W ]
        
        self.matrix_fem = Convection(self.Space.fem,
                                     u_conv,assemble='yes',
                                     silent=self.silent).matrix
        #return Convection(self.Space.fem,u_conv,assemble='yes').matrix
    
    def __action_1d(self,v_in):
        # Mapping from global to local
        v = self.Space.mapping_q(v_in)
        # Matrix-vector product
        w = operators_f90.action_of_convection_1d(v,self.u_conv,self.Space.weights,\
                                                  self.Space.D)
        # Mapping from local to global
        v_out=self.Space.mapping_qt(w)
        # Mask
        v_out = self.Space.mask(v_out)
        return v_out
        
    def __action_2d(self,v_in):
        # Mapping from global to local
        v = self.Space.mapping_q(v_in)
        # Matrix-vector product
        w = operators_f90.action_of_convection_2d(v,self.u_conv[0],\
                                                  self.u_conv[1],\
                                                  self.Space.weights,\
                                                  self.Space.D,\
                                                  self.g11,self.g12,\
                                                  self.g21,self.g22)    
        # Mapping from local to global
        v_out=self.Space.mapping_qt(w)
        # Mask
        v_out = self.Space.mask(v_out)
        return v_out
    
    def __action_3d(self,v_in):
        # Mapping from global to local
        v = self.Space.mapping_q(v_in)
        # Matrix-vector product
        #w = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),float)
        #for k in range(self.Space.noe):
        #    w[k,:,:,:] = operators_f90.action_of_convection_3d(v[k,:,:,:],\
        #                           self.u_conv[0][k,:,:,:],self.u_conv[1][k,:,:,:],self.u_conv[2][k,:,:,:],\
        #                           self.W,self.Space.D,\
        #                           self.g11[k,:,:,:],self.g12[k,:,:,:],self.g13[k,:,:,:],\
        #                           self.g21[k,:,:,:],self.g22[k,:,:,:],self.g23[k,:,:,:],\
        #                           self.g31[k,:,:,:],self.g32[k,:,:,:],self.g33[k,:,:,:])
        w = operators_f90.action_of_convection_3d_2(v,\
                                   self.u_conv[0],self.u_conv[1],self.u_conv[2],\
                                   self.W,self.Space.D,\
                                   self.g11,self.g12,self.g13,\
                                   self.g21,self.g22,self.g23,\
                                   self.g31,self.g32,self.g33)
        # Mapping from local to global
        v_out=self.Space.mapping_qt(w)
        # Mask
        v_out = self.Space.mask(v_out)
        return v_out
        
    def __assemble_1d(self):
        '''
        Assemble.
        '''
        if not self.silent:
            print ''
            print '**** Assemble convection      ****'
        nz = self.Space.noe * self.Space.n * self.Space.n
        t1 = time.clock()
        irow,jcolumn,value = \
            operators_f90.assemble_convection_1d(self.u_conv,self.Space.weights,self.Space.D,\
                                                self.Space.theta,nz)
        #A = pysparse.spmatrix.ll_mat(self.Space.dof, self.Space.dof, nz)
        #A = sps.lil_matrix((self.Space.dof, self.Space.dof), dtype='float')
        #A = pysparseMatrix.PysparseMatrix(size=self.Space.dof)
        A = PysparseMatrix(size=self.Space.dof)
        # Assign values
        for k in range(len(value)):
            i = int(irow[k])
            j = int(jcolumn[k])
            A[i,j] = A[i,j] + value[k]
        # Mask
        A = self.Space.mask_matrix(A)
        t2 = time.clock()
        if not self.silent:
            print '**** ex time=',t2-t1,'nnz=',A.nnz
        return A
        
    def __assemble_2d(self):
        # Local comnectivity array
        theta_local = np.zeros((self.Space.n, self.Space.n),int)
        count =0
        for m in range(self.Space.n):
            for n in range(self.Space.n):
                theta_local[m,n]=count
                count=count+1

        # Mask array
        mask   = np.zeros((self.Space.noe,self.Space.n,self.Space.n),int)
        mask_g = np.zeros((self.Space.dof),int)
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                [i, phys, bc_type, edge] = self.Space.theta_phys_bc(k,m)
                if bc_type == "Dir":
                    mask_g[i] = 1
                        
        for k in range(self.Space.noe):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    i = self.Space.theta[k,m,n]
                    mask[k,m,n] = mask_g[i]
        del mask_g
        
        if not self.silent:
            print ' '
            print '**** Assemble convection      ****'
        # Stiffness matrix
        #A  = pysparseMatrix.PysparseMatrix(size=self.Space.dof)#
        A  = PysparseMatrix(size=self.Space.dof)#
        nz = self.Space.n*self.Space.n
        t1 = time.clock()
        # Assemble the stiffness matrix elementwize.
        for k in range(self.Space.noe):
            val,row,col = \
                   operators_f90.assemble_convection_2d(self.Space.weights,\
                        self.u_conv[0][k,:,:],self.u_conv[1][k,:,:],\
                        self.g11[k,:,:],self.g12[k,:,:],\
                        self.g21[k,:,:],self.g22[k,:,:],\
                        self.Space.D,self.Space.theta[k,:,:],theta_local,mask[k,:,:],nz)
            val = np.reshape(val, nz*nz)
            row = np.reshape(row, nz*nz)
            col = np.reshape(col, nz*nz)
            A.addAt(val,row,col)
        #---
        # Mask
        #t1_m = time.clock()
        #A = self.Space.mask_matrix(A)
        #t2_m = time.clock()
               
        t2 = time.clock()
        #print '**** mask     time=',t2_m-t1_m
        if not self.silent:
            print '**** assemble time=',t2-t1
            print '**** nnz=',A.nnz
        return A
        
    def __assemble_3d(self):
        # Local comnectivity array
        theta_local = np.zeros((self.Space.n, self.Space.n, self.Space.n),int)
        count =0
        for k in range(self.Space.n):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    theta_local[k,m,n]=count
                    count=count+1

        # Mask array
        mask   = np.zeros((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),int)
        mask_g = np.zeros((self.Space.dof),int)
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    [i, phys, bc_type, edge] = self.Space.theta_phys_bc(k,m,n)
                    if bc_type == "Dir":
                        mask_g[i] = 1
                        
        for k in range(self.Space.noe):
            for m in range(self.Space.n):
                for n in range(self.Space.n):
                    for o in range(self.Space.n):
                        i = self.Space.theta[k,m,n,o]
                        mask[k,m,n,o] = mask_g[i]
        del mask_g
        if not self.silent:
            print ''
            print '**** Assemble Convection      ****'
        # Matrix
        #A = pysparseMatrix.PysparseMatrix(size=self.Space.dof)#, sizeHint=nz)
        A = PysparseMatrix(size=self.Space.dof)
        
        # Local degrees of freedom
        nz =  self.Space.n*self.Space.n*self.Space.n
        t1 = time.clock()
        
        # Assemble the stiffness matrix, elementwize.
        t1_f = time.clock()    
        for k in range(self.Space.noe):
            val,row,col = \
                operators_f90.assemble_convection_3d(self.Space.weights,\
                          self.u_conv[0][k,:,:,:],self.u_conv[1][k,:,:,:],self.u_conv[2][k,:,:,:],\
                          self.g11[k,:,:,:],self.g12[k,:,:,:],self.g13[k,:,:,:],\
                          self.g21[k,:,:,:],self.g22[k,:,:,:],self.g23[k,:,:,:],\
                          self.g31[k,:,:,:],self.g32[k,:,:,:],self.g33[k,:,:,:],\
                          self.Space.D,self.Space.theta[k,:,:,:],theta_local,mask[k,:,:,:],nz)
            val = np.reshape(val, nz*nz)
            row = np.reshape(row, nz*nz)
            col = np.reshape(col, nz*nz)
            A.addAt(val,row,col)
            
        t2_f = time.clock()    
        
        t2 = time.clock()
        
        # Mask
        if not self.silent:
            print '**** Number of nnz=',A.nnz
            #print '**** post mask nnz=',A.nnz
            #print '**** assemble time=',t2_f-t1_f
            #print '**** robin    time=',t2_r-t1_r
            #print '**** mask     time=',t2_m-t1_m
            print '**** assemble time=',t2-t1#,'nnz=',A.nnz
        
        return A


class MultipleOperators:
    '''
    Combines multiple operators into one, e.g. 
    
    .. math::
    
       (A+B+C)u=b
    
    becomes 
    
    .. math::
    
       Lu=b
    
    where :math:`L=A+B+C`. 
    
    :param list Operator: A list of operators
    :param list scaling_factor: Scaling factors.
    :param assemble: A flag which determines if the combined operator should
                     be assembled (:literal:`assemble='yes'`) or not
                     (:literal:`assemble='no'`).
    :type assemble: string
    :attribute: * **matrix** (*Scipy LinearOperator/Pysparse matrix*)- The combined operator. 
           
    **Example**::
    
      import sempy
      
      X = sempy.Space( filename = 'square', n = 4, dim = 2 )
      
      lap = sempy.operators.Laplacian( X, fem_assemble = 'yes' )
      A_fem = lap.matrix_fem
      
      conv = sempy.operators.Convection( X, fem_assemble = 'yes' )
      C_fem = conv.matrix_fem

      # L = - A_fem - 2.0 * C_fem
      combined = sempy.operators.MultipleOperators( [A_fem,C_fem],
                                                    [-1,-2],
                                                    assemble = 'no' )
      L = combined.matrix
    
    '''
    def __init__(self,Operator,scaling_factor='none',assemble='no',
                 silent=False):
        self.Operator = Operator
        self.dof      = self.Operator[0].shape[0]
        self.silent = silent
        
        if scaling_factor == 'none':
            self.scaling_factor = np.ones((len(self.Operator)),float)
        else:
            self.scaling_factor = scaling_factor
        
        if assemble == 'no':
            self.matrix = spl.LinearOperator( (self.dof,self.dof),\
                                   matvec=self.__action__,dtype='float')
        if assemble == 'yes':
            if not self.silent:
                print ''
                print '**** MultipleOperators        ****'
            t1 = time.clock()
            #self.matrix = pysparseMatrix.PysparseMatrix(size=self.dof)
            self.matrix = PysparseMatrix(size=self.dof)
            for i in range(len(self.Operator)):
                val,irow,jcol = self.Operator[i].find()
                factor = float(self.scaling_factor[i])
                self.matrix.addAt(val*factor, irow, jcol)
            self.matrix.dtype='float64'
            t2 = time.clock()
            if not self.silent:
                print '**** Number of nnz=',self.matrix.nnz
                print '**** Assemble time =',t2-t1
            
    def __action__(self,v):
        w = np.zeros((self.dof))
        for i in range(len(self.Operator)):
            factor = float(self.scaling_factor[i])
            s = self.Operator[i] * v
            w = w + factor * s
            #del s
        return w   


       
class Mass:
    '''
    The mass matrix.
    
    .. math::
       \int_\Omega uv\, d\Omega\Rightarrow Mu
    
    :param Space: The discrete function space. 
    :type Space: :class:`sempy.Space`
    :param mask: If :literal:`mask='yes'` the matrix is masked according to 
                 strong boundary conditions. If :literal:`mask='no'` the
                 matrix is not masked.
    :type mask: string
    
    :attributes: * **matrix** - The mass matrix.
                 * **matrix_inv** - The inverse of the mass matrix. 
                 
    **Example**::
    
      import sempy
      
      
      X = sempy.Space( 'square', n = 7, dim = 2 )

      mass = sempy.operators.Mass(X)
      M     = mass.matrix
      M_inv = mass.matrix_inv
      
      f = sempy.Function( X, basis_coeff = 2.0 )
      b = mass.action_local( f.basis_coeff )
      
    '''
    def __init__( self, Space, mask='yes' ):
        self.Space = Space
        self.mask = mask
        self.matrix = self.__assemble_mass()
        self.matrix_inv = self.__assemble_mass_inv()
    
      
    def __assemble_mass(self):
        # The local matrix vector product w=M_L*u_L
        if self.Space.dim == 1:
            u_local = np.ones((self.Space.noe,self.Space.n),float)
            w = operators_f90.action_of_mass_1d(u_local,
                                                self.Space.jac,
                                                self.Space.weights)
        if self.Space.dim == 2:
            u_local = np.ones((self.Space.noe,self.Space.n,self.Space.n),float)
            w = operators_f90.action_of_mass_2d(u_local,
                                                self.Space.jac,
                                                self.Space.weights)
        if self.Space.dim == 3:
            u_local = np.ones((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),float)
            w = operators_f90.action_of_mass_3d(u_local,
                                                self.Space.jac,
                                                self.Space.weights)
        # Gather (local to global)
        w_out = self.Space.mapping_qt(w)
        # Mask
        if self.mask == 'yes':
            w_out = self.Space.mask(w_out)
        # Matrix
        #A = pysparseMatrix.PysparseMatrix(size = self.Space.dof)
        A = PysparseMatrix(size = self.Space.dof)
        A.putDiagonal(w_out)
        return A
    
    def __assemble_mass_inv(self):
        if self.Space.dim == 1:
            u_local = np.ones((self.Space.noe,self.Space.n),float)
            w = operators_f90.action_of_mass_1d( u_local,
                                                 self.Space.jac,
                                                 self.Space.weights)
        if self.Space.dim == 2:
            u_local = np.ones( (self.Space.noe,self.Space.n,self.Space.n),
                               float )
            w = operators_f90.action_of_mass_2d( u_local, 
                                                 self.Space.jac,
                                                 self.Space.weights)
        if self.Space.dim == 3:
            u_local = np.ones((self.Space.noe,self.Space.n,self.Space.n,self.Space.n),float)
            w = operators_f90.action_of_mass_3d(u_local,self.Space.jac,self.Space.weights)
        # Gather (local to global)
        w_out = self.Space.mapping_qt(w)
        # Inverse
        if self.Space.dim == 1:
            w_out = operators_f90.mass_inv_1d(w_out)
        if self.Space.dim == 2:
            w_out = operators_f90.mass_inv_2d(w_out)
        if self.Space.dim == 3:
            w_out = operators_f90.mass_inv_3d(w_out)
        # Mask
        if self.mask == 'yes':
            w_out = self.Space.mask(w_out)
        # Matrix
        #A = pysparseMatrix.PysparseMatrix(size = self.Space.dof)
        A = PysparseMatrix(size = self.Space.dof)
        A.putDiagonal(w_out)
        return A
        
    def action_local(self,u_L):
        '''
        Action of the mass matrix on a set of basis coefficients 
        in local data structure:
        
        .. math::
           v=Mu=Q^TM_Lu_L
           
        :param u_L: Basis coefficients in local data structure.
        :returns: A vector in :math:`\mathbb R^{dof}`
           
        '''
        # The local matrix vector product w=M_L*u_L
        if self.Space.dim == 1:
            w = operators_f90.action_of_mass_1d(u_L,self.Space.jac,self.Space.weights)
        if self.Space.dim == 2:
            w = operators_f90.action_of_mass_2d(u_L,self.Space.jac,self.Space.weights)
        if self.Space.dim == 3:
            w = operators_f90.action_of_mass_3d(u_L,self.Space.jac,self.Space.weights)
        # Gather (local to global)
        w_out = self.Space.mapping_qt(w)
        # Mask
        if self.mask == 'yes':
            w_out = self.Space.mask(w_out)
            
        return w_out
        



import sempy
import numpy as np

import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.mplot3d import axes3d
from matplotlib import rcParams
rcParams['text.usetex']=True
rcParams['text.latex.unicode']=True
rcParams['legend.fontsize']=20

from pysparse.sparse.pysparseMatrix import PysparseMatrix


class Diffusion:
    '''
    Test case for the diffusion equation
    
    .. math::
    
       y=
       
    '''
    def __init__(self,Space,mu=None):
        self.Space = Space
        self.M = sempy.operators.Mass( self.Space ).matrix
        self.mu = mu
        self.f = sempy.Function( self.Space)
        if self.mu == None:
            self.mu = sempy.Function( self.Space, basis_coeff = 1.0)
        self.error = sempy.Function( self.Space )
        self.u = sempy.Function( self.Space)
        
    def initial_condition(self,y,t):
        #
        if self.Space.dim == 1:
            y_val = np.cos(t)*np.cos(2.0*np.pi*self.Space.x)
            
        if self.Space.dim == 2:
            y_val = np.cos(t)*np.cos(2.0*np.pi*self.Space.x)*\
                              np.cos(2.0*np.pi*self.Space.y)
        y.basis_coeff = np.copy( y_val )
        return y
    
    def force_function(self,y,t):
        #
        if self.Space.dim == 1:
            dudt  = -np.sin(t)*np.cos(2.0*np.pi*self.Space.x)
            dudx2 = -4.0*np.pi*np.pi*\
                     np.cos(t)*np.cos(2.0*np.pi*self.Space.x)
            self.f.basis_coeff = dudt - self.mu.basis_coeff*( dudx2  )
            
        if self.Space.dim == 2:
            dudt  = -np.sin(t)*np.cos(2.0*np.pi*self.Space.x)*\
                               np.cos(2.0*np.pi*self.Space.y)
            dudx2 = -4.0*np.pi*np.pi*np.cos(t)*\
                     np.cos(2.0*np.pi*self.Space.x)*\
                     np.cos(2.0*np.pi*self.Space.y)
            self.f.basis_coeff = dudt - self.mu.basis_coeff*( 2.0*dudx2  )
                
        b = self.M*self.f.glob()
        return [b]
        
    def bc(self,y,t):
        if self.Space.dim == 1:
            y_val = np.cos(t)*np.cos(2.0*np.pi*self.Space.x)
                
        if self.Space.dim == 2:
            y_val = np.cos(t)*np.cos(2.0*np.pi*self.Space.x)*\
                              np.cos(2.0*np.pi*self.Space.y)
        
        self.u.basis_coeff = self.Space.mapping_q( y )
        for i in range(1,5):
            self.u.set_bc(y_val,i)
            
        return self.u.glob()
    
    def error_norms( self, y, t ):
        # Error norms.
        y_ex = sempy.Function( self.Space )
        y_ex = self.initial_condition( y_ex, t )
        
        self.error.basis_coeff = y_ex.basis_coeff - y.basis_coeff
        e_L2=self.error.l2_norm()
        e_H1=self.error.h1_norm()
        print 'L2= ',e_L2
        print 'H1= ',e_H1  
        return e_L2,e_H1
    
    def plot_error(self):
        self.error.plot_wire()


class ConvectionDiffusion1d:
    '''
    Test case for the convection-diffusion equation.
    '''
    def __init__(self,Space,mu=None):
        self.Space = Space
        self.M = sempy.operators.Mass( self.Space ).matrix
        self.mu = mu
        self.f = sempy.Function( self.Space)
        if self.mu == None:
            self.mu = sempy.Function( self.Space, basis_coeff = 1.0)
        self.error = sempy.Function( self.Space )
        self.u = sempy.Function( self.Space)
        self.C = sempy.operators.Convection( self.Space ).matrix
            
    def initial_condition(self,y,t):
        y_val = np.exp(-4.0*np.pi*np.pi*self.mu.basis_coeff*t)*\
                np.sin(2.0*np.pi*(self.Space.x-t))
        y.basis_coeff = np.copy( y_val )
        return y
    
    def force_function(self,y,t):
        #dudt  = -np.sin(t)*\
        #         np.cos(2.0*np.pi*self.Space.x)#*\
        #         #np.cos(2.0*np.pi*self.Space.y)
        #dudx2 = -4.0*np.pi*np.pi*\
        #             np.cos(t)*\
        #             np.cos(2.0*np.pi*self.Space.x)#*\
        #             #np.cos(2.0*np.pi*self.Space.y)
        #self.f.basis_coeff = dudt - self.mu.basis_coeff*( 1.0*dudx2  )
        #b = self.M*self.f.glob()
        #b=np.zeros((self.Space.dof),float)
        b =  self.C * y[0].glob()
        b = -b
        return [b]
        
    def bc(self,y,t):
        #
        y_val = np.exp(-4.0*np.pi*np.pi*self.mu.basis_coeff*t)*\
                np.sin(2.0*np.pi*(self.Space.x-t))
        #self.u.basis_coeff = self.Space.mapping_q( y )
        for i in range(0,3):
            #self.u.set_bc(y_val,i)
            y[0].set_bc(y_val,i)
        return y#y[0]#self.u.glob()
    
    def error_norms( self, y, t ):
        # Error norms.
        y_ex = sempy.Function( self.Space )
        y_ex = self.initial_condition( y_ex, t )
        
        self.error.basis_coeff = y_ex.basis_coeff - y.basis_coeff
        e_L2=self.error.l2_norm()
        e_H1=self.error.h1_norm()
        print 'L2= ',e_L2
        print 'H1= ',e_H1  
        return e_L2,e_H1
    
    def plot_error(self):
        self.error.plot_wire()

class ConvectionDiffusion2d:
    '''
    Test case for the convection-diffusion equation.
    '''
    def __init__(self,Space,mu=None):
        self.Space = Space
        self.M = sempy.operators.Mass( self.Space ).matrix
        self.mu = mu
        self.f = sempy.Function( self.Space)
        if self.mu == None:
            self.mu = sempy.Function( self.Space, basis_coeff = 1.0)
        self.error = sempy.Function( self.Space )
        self.u = sempy.Function( self.Space)
        self.C = sempy.operators.Convection( self.Space ).matrix
            
    def initial_condition(self,u,t):
        x=self.Space.x
        y=self.Space.y
        u_val = -np.sin(np.pi*x)*np.cos(np.pi*y)*np.cos(np.pi*t)
        u.basis_coeff = np.copy( u_val )
        return u
    
    def force_function(self,y,t):
        #dudt  = -np.sin(t)*\
        #         np.cos(2.0*np.pi*self.Space.x)#*\
        #         #np.cos(2.0*np.pi*self.Space.y)
        #dudx2 = -4.0*np.pi*np.pi*\
        #             np.cos(t)*\
        #             np.cos(2.0*np.pi*self.Space.x)#*\
        #             #np.cos(2.0*np.pi*self.Space.y)
        #self.f.basis_coeff = dudt - self.mu.basis_coeff*( 1.0*dudx2  )
        #b = self.M*self.f.glob()
        #b=np.zeros((self.Space.dof),float)
        b =  self.C * y
        b = -b
        return [b]
        
    def bc(self,u,t):
        #
        #y_val = np.exp(-4.0*np.pi*np.pi*self.mu.basis_coeff*t)*\
        #        np.sin(2.0*np.pi*(self.Space.x-t))
        x=self.Space.x
        y=self.Space.y
        u_val = -np.sin(np.pi*x)*np.cos(np.pi*y)*np.cos(np.pi*t)
        #self.u.basis_coeff = self.Space.mapping_q( u_val )
        for i in range(0,3):
            self.u.set_bc(u_val,i)
        return self.u.glob()
    
    def error_norms( self, y, t ):
        # Error norms.
        y_ex = sempy.Function( self.Space )
        y_ex = self.initial_condition( y_ex, t )
        
        self.error.basis_coeff = y_ex.basis_coeff - y.basis_coeff
        e_L2=self.error.l2_norm()
        e_H1=self.error.h1_norm()
        print 'L2= ',e_L2
        print 'H1= ',e_H1  
        return e_L2,e_H1
    
    def plot_error(self):
        self.error.plot_wire()   


class TwoEqTest:
    '''
    Test case for the convection-diffusion equation.
    '''
    def __init__(self,Space,mu_1=None,mu_2=None):
        self.Space = Space
        self.M = sempy.operators.Mass( self.Space ).matrix
        self.mu_1 = mu_1
        self.mu_2 = mu_2
        if self.mu_1 == None:
            self.mu_1 = sempy.Function( self.Space, basis_coeff = 1.0)
        if self.mu_2 == None:
            self.mu_2 = sempy.Function( self.Space, basis_coeff = 1.0)
        self.f = sempy.Function( self.Space)
        self.error = [sempy.Function(self.Space),sempy.Function(self.Space)]
        #self.u = sempy.Function( self.Space)
        self.C = sempy.operators.Convection( self.Space ).matrix
        
    def linear_operator(self):
        lap_1 = sempy.operators.Laplacian( self.Space,fem_assemble='yes',
                                         mu=self.mu_1 )
        A_1 = lap_1.matrix
        A_1_fem= lap_1.matrix_fem
        
        lap_2 = sempy.operators.Laplacian( self.Space,fem_assemble='yes',
                                         mu=self.mu_2 )
        A_2 = lap_2.matrix
        A_2_fem= lap_2.matrix_fem
        
        B_1 = sempy.operators.MultipleOperators([A_1], [-1.0 ]).matrix
        B_1_fem = sempy.operators.MultipleOperators([A_1_fem], 
                                           [-1.0],
                                           assemble='yes').matrix
        B_2 = sempy.operators.MultipleOperators([A_2], [-1.0 ]).matrix
        B_2_fem = sempy.operators.MultipleOperators([A_2_fem], 
                                           [-1.0],
                                           assemble='yes').matrix
        E = PysparseMatrix(size=self.Space.dof)
                                    
        S = [[B_1, E],[E,   B_2]]
        S_fem = [[B_1_fem,E],[E,B_2_fem]]
        return S,S_fem
            
    def initial_condition(self,y,t):
        # Conv-diff eq
        y_val = np.exp(-4.0*np.pi*np.pi*self.mu_1.basis_coeff*t)*\
                np.sin(2.0*np.pi*(self.Space.x-t))
        y[0].basis_coeff = np.copy( y_val )
        # Diff eq
        y_val = np.cos(t)*np.cos(2.0*np.pi*self.Space.x)
        y[1].basis_coeff = np.copy( y_val )
        return y
    
    def force_function(self,y,t):
        # First
        b_1 =  self.C * y[0].glob()
        b_1 = -b_1
        # Second
        dudt  = -np.sin(t)*np.cos(2.0*np.pi*self.Space.x)
        dudx2 = -4.0*np.pi*np.pi*\
               np.cos(t)*np.cos(2.0*np.pi*self.Space.x)
        self.f.basis_coeff = dudt - self.mu_2.basis_coeff*( dudx2  )
        b_2 = self.M*self.f.glob()
        return [b_1,b_2]
        
    def bc(self,y,t):
        # First
        y_val = np.exp(-4.0*np.pi*np.pi*self.mu_1.basis_coeff*t)*\
                np.sin(2.0*np.pi*(self.Space.x-t))
        for i in range(0,3):
            y[0].set_bc(y_val,i)
        # Second
        y_val = np.cos(t)*np.cos(2.0*np.pi*self.Space.x)
        for i in range(0,3):
            y[1].set_bc(y_val,i)
        return y
    
    def error_norms( self, y, t ):
        # Error norms.
        y_ex = [sempy.Function(self.Space),sempy.Function(self.Space)]
        y_ex = self.initial_condition( y_ex, t )
        
        self.error[0].basis_coeff = y_ex[0].basis_coeff - y[0].basis_coeff
        self.error[1].basis_coeff = y_ex[1].basis_coeff - y[1].basis_coeff

        e1_L2 = self.error[0].l2_norm()
        e1_H1 = self.error[0].h1_norm()
        e2_L2 = self.error[1].l2_norm()
        e2_H1 = self.error[1].h1_norm()
        
        return e1_L2,e1_H1,e2_L2,e2_H1
    
    def plot_error(self,i):
        self.error[i].plot_wire()


class TwoEqHeat:
    '''
    Test case for the two-equation heat transfer model.
    '''
    def __init__(self,Space,gamma=None,eta=None,Pe=1.0,Time=None):
        self.Space = Space
        self.M = sempy.operators.Mass( self.Space ).matrix
        self.gamma = gamma
        self.eta = eta
        if self.gamma == None:
            self.gamma = sempy.Function( self.Space, basis_coeff = 1.0)
        if self.eta == None:
            self.eta = 1.0
        self.f = sempy.Function( self.Space)
        self.error = [sempy.Function(self.Space),sempy.Function(self.Space)]
        self.Pe=Pe
        u_conv = sempy.Function( self.Space, basis_coeff = self.Pe)
        self.conv = sempy.operators.Convection( self.Space, 
                                             u_conv=[u_conv] )
        self.C = self.conv.matrix
        # Plot of solution
        self.Time = Time
        if not self.Time==None:
            n_t = self.Time.time_steps+1
            t = np.zeros((n_t),float)
            h = self.Time.h
            for i in range(n_t):
                t[i]=np.float(i)*h
            #v = np.zeros((n_t,u.Space.dof),float)
            x = self.Space.local_to_global(self.Space.x)
            self.X=np.zeros((n_t,self.Space.dof),float)
            self.T=np.zeros((n_t,self.Space.dof),float)
            self.y1=np.zeros((n_t,self.Space.dof),float)
            self.y2=np.zeros((n_t,self.Space.dof),float)
            for i in range(n_t):
                self.X[i,:]=x[:]
                self.T[i,:]=t[i]
            
            
    def linear_operator(self):
        '''
        Linear operator
        '''
        # L_11
        lap_1 = sempy.operators.Laplacian( self.Space,fem_assemble='yes',
                                            mu=self.gamma )
        
        A_1 = lap_1.matrix
        A_1_fem= lap_1.matrix_fem
                
        L_11 = sempy.operators.MultipleOperators([A_1,self.M], 
                                                 [-1.0,-self.eta]).matrix
        L_11_fem = sempy.operators.MultipleOperators([A_1_fem,self.M], 
                                           [-1.0,-self.eta],
                                           assemble='yes').matrix
        # L_12
        L_12 = sempy.operators.MultipleOperators([self.M], [self.eta],
                                                 assemble='yes').matrix
        # L_21
        L_21 = sempy.operators.MultipleOperators([self.M], [self.eta],
                                                 assemble='yes').matrix
        # L_22
        lap_2 = sempy.operators.Laplacian(self.Space,fem_assemble='yes')
        A_2 = lap_2.matrix
        A_2_fem= lap_2.matrix_fem
        L_22 = sempy.operators.MultipleOperators([A_2,self.M], 
                                                 [-1.0,-self.eta]).matrix
        L_22_fem = sempy.operators.MultipleOperators([A_2_fem,self.M], 
                                           [-1.0,-self.eta],
                                           assemble='yes').matrix
        #                                    
        S = [[L_11, L_12],[L_21, L_22]]
        S_fem = [[L_11_fem,L_12],[L_21,L_22_fem]]
        return S,S_fem
            
    def initial_condition(self,y,t):
        # First
        x=self.Space.x
        y_val = np.exp(-t)*np.sin(2.0*np.pi*x)
        y[0].basis_coeff[:,:] = np.copy( y_val )
        # Second
        g1=(4.0*self.gamma.basis_coeff)*np.pi*np.pi + self.eta - 1.0
        g=(1.0/self.eta)*(g1*np.sin(2.0*np.pi*x)+\
                          2.0*np.pi*self.Pe*np.cos(2.0*np.pi*x))
        y_val=np.exp(-t)*g
        y[1].basis_coeff[:,:] = np.copy( y_val )
        return y
    
    def force_function(self,y,t):
        # First
        b_1 =  self.C * y[0].glob()
        b_1 = -b_1
        # Second
        x=self.Space.x
        g1=(4.0*self.gamma.basis_coeff)*np.pi*np.pi + self.eta - 1.0
        g=(1.0/self.eta)*(g1*np.sin(2.0*np.pi*x)+\
                          2.0*np.pi*self.Pe*np.cos(2.0*np.pi*x))
        Q=(4.0*np.pi*np.pi + self.eta - 1.0)*g - self.eta*np.sin(2.0*np.pi*x)
        f_val = np.exp(-t)*Q
        
        self.f.basis_coeff[:,:] = np.copy(f_val)
        b_2 = self.M*self.f.glob()
        return [b_1,b_2]
    
    def callback(self,y,t,i):
        # First
        y_ex = [sempy.Function(self.Space),sempy.Function(self.Space)]
        y_ex = self.initial_condition( y_ex, t )
        
        self.error[0].basis_coeff = y_ex[0].basis_coeff - y[0].basis_coeff
        self.error[1].basis_coeff = y_ex[1].basis_coeff - y[1].basis_coeff
        
        #self.y1[i,:]=y[0].glob()
        #self.y2[i,:]=y[1].glob()
        self.y1[i,:]=self.error[0].glob()
        self.y2[i,:]=self.error[1].glob()
        return
    
    def plot_evo(self):
        font_size = 20
        
        fig1 = plt.figure(1)
        ax1 = axes3d.Axes3D(fig1)
        #ax1.plot_wireframe(self.X,self.T, self.y1)
        #ax1.plot_surface(self.X,self.T, self.y1,
        #                 rstride=1, cstride=1, cmap=cm.jet,
        #                 linewidth=1, antialiased=False)
        ax1.plot_surface( self.X,self.T,self.y1,
                          color='b',alpha=1.0,rstride=1, cstride=1)
        # Dirk252, oifs2
        #ax1.set_zlim3d(-0.004,0.008)
        # Strang
        #ax1.set_zlim3d(-0.015,0.015)
        for label in ax1.w_xaxis.get_ticklabels():
            label.set_fontsize(18)
        for label in ax1.w_yaxis.get_ticklabels():
            label.set_fontsize(18)
        for label in ax1.w_zaxis.get_ticklabels():
            label.set_fontsize(18)
        ax1.set_xlabel(r'$x$',fontsize=font_size)
        ax1.set_ylabel(r'$t$',fontsize=font_size)
        ax1.set_zlabel(r'$e$',fontsize=font_size)
        
        fig2 = plt.figure(2)
        ax2 = axes3d.Axes3D(fig2)
        ax2.plot_surface( self.X,self.T,self.y2,
                          color='b',alpha=1.0,rstride=1, cstride=1)
        #ax2.set_zlim3d(-0.00006,0.00006)

        for label in ax2.w_xaxis.get_ticklabels():
            label.set_fontsize(12)
        for label in ax2.w_yaxis.get_ticklabels():
            label.set_fontsize(12)
        for label in ax2.w_zaxis.get_ticklabels():
            label.set_fontsize(12)
                
        ax2.set_xlabel(r'$x$',fontsize=font_size)
        ax2.set_ylabel(r'$t$',fontsize=font_size)
        ax2.set_zlabel(r'$e$',fontsize=font_size)
        
        
        plt.show()   
        
        
        
    def bc(self,y,t):
        # First
        #y_val = np.exp(-4.0*np.pi*np.pi*self.mu_1.basis_coeff*t)*\
        #        np.sin(2.0*np.pi*(self.Space.x-t))
        self.error = self.initial_condition( self.error, t )
        #y_val=0.0
        for i in range(0,3):
            y[0].set_bc(self.error[0].basis_coeff,i)
        # Second
        #y_val = np.cos(t)*np.cos(2.0*np.pi*self.Space.x)
        #y_val=0.0
        for i in range(0,3):
            y[1].set_bc(self.error[1].basis_coeff,i)
        return y
    
    def error_norms( self, y, t ):
        # Error norms.
        y_ex = [sempy.Function(self.Space),sempy.Function(self.Space)]
        y_ex = self.initial_condition( y_ex, t )
        
        self.error[0].basis_coeff = y_ex[0].basis_coeff - y[0].basis_coeff
        self.error[1].basis_coeff = y_ex[1].basis_coeff - y[1].basis_coeff

        e1_L2 = self.error[0].l2_norm()
        e1_H1 = self.error[0].h1_norm()
        e2_L2 = self.error[1].l2_norm()
        e2_H1 = self.error[1].h1_norm()
        
        return e1_L2,e1_H1,e2_L2,e2_H1
    
    def plot_error(self):
        self.error[0].plot_wire(show=False)
        self.error[1].plot_wire(show=True)


'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 

import numpy as np
import sempy
#import sempy.basic as basic
#import sempy.space.space_f90 as space_f90

#import matplotlib.pyplot as plt
from matplotlib import pyplot as pl
#from mpl_toolkits.mplot3d import axes3d
#from matplotlib import cm


class Plot:
    '''
    An interface to matplotlib for easy plotting of 
    Sempy functions
    '''
    def __init__(self,Space):
        self.Space = Space
    
    def plot3d(self, u,contours=[0.01,0.02]):
        try:
            import enthought.mayavi
        except:
            print 'No mayavi python module installed.'
            print '3D plotting not possible.'
            return

        v_fem = sempy.Function(self.Space.fem, filename = 'test_data')
        v_fem.basis_coeff = self.Space.fem.sem_to_fem(u.basis_coeff)
        v_fem.to_file()

        vtkFile='test_data0.vtk'

        # Create the MayaVi engine and start it.
        from enthought.mayavi.core.api import Engine
        engine = Engine()
        engine.start()
        scene = engine.new_scene()

        # Read in VTK file and add as source
        from enthought.mayavi.sources.vtk_file_reader import VTKFileReader
        reader = VTKFileReader()
        reader.initialize(vtkFile)
        engine.add_source(reader)

        ## Add Surface Module
        #from enthought.mayavi.modules.surface import Surface
        #surface = Surface()
        #engine.add_module(surface)

        # Axes
        from enthought.mayavi.modules.axes import Axes
        axes = Axes()
        engine.add_module(axes)

        # Outline
        from enthought.mayavi.modules.outline import Outline
        outline = Outline()
        engine.add_module(outline)

        # IsoSurface
        from enthought.mayavi.modules.api import IsoSurface
        iso = IsoSurface()
        engine.add_module(iso)
        #iso.contour.contours = [-0.4, 0.0, 0.4]
        #iso.contour.contours = [-0.9, 0.45, 0.0, 0.45, 0.9]
        iso.contour.contours = contours#[0.01, 0.02, 0.03]
        # Make them translucent.
        #iso.actor.property.opacity = 0.9
        # Show the scalar bar (legend).
        iso.module_manager.scalar_lut_manager.show_scalar_bar = True

        # Move the camera
        scene.scene.camera.elevation(-45)

        # Save scene to image file
        #scene.scene.save_png('figures/poisson_3d_plot.png')

        # Create a GUI instance and start the event loop.
        # This stops the window from closing
        from enthought.pyface.api import GUI
        gui = GUI()
        gui.start_event_loop()
            
    def plot2d( self, u, min = 0.0, max = 0.1, levels = 10,
                n_int = 10, type = 'countorf' ):
        '''
        Plot a 2D function.
        '''
        fig = pl.figure()
        x = self.Space.interpolation(self.Space.x,n_int)
        y = self.Space.interpolation(self.Space.y,n_int)
        z = self.Space.interpolation(u.basis_coeff,n_int)
        
        du=(max-min)/float(levels)
        levels = np.arange(min, max, du)
        
        if type == 'countorf':
            for k in range(self.Space.noe):
                cset1 = pl.contourf(x[k,:,:], y[k,:,:], z[k,:,:], levels)
        if type == 'countor':
            for k in range(self.Space.noe):
                cset1 = pl.contour(x[k,:,:], y[k,:,:], z[k,:,:], levels)
        pl.xlabel(r'$x$')
        pl.ylabel(r'$y$')              
        pl.axis([self.Space.x.min(), self.Space.x.max(), self.Space.y.min(), self.Space.y.max()])  
        pl.colorbar(cset1)
        pl.show()

        